import pytest
from vcr.serializers.jsonserializer import serialize
from vcr.request import Request


def test_serialize_binary():
    request = Request(
        method='GET',
        uri='http://localhost/',
        body='',
        headers={},
    )
    cassette = {'requests': [request], 'responses': [{'body': b'\x8c'}]}

    with pytest.raises(Exception) as e:
        serialize(cassette)
        assert e.message == "Error serializing cassette to JSON. Does this \
            HTTP interaction contain binary data? If so, use a different \
            serializer (like the yaml serializer) for this request"



# -*- coding: utf-8 -*-
'''Basic tests for cassettes'''

# External imports
import os
from six.moves.urllib.request import urlopen

# Internal imports
import vcr


def test_nonexistent_directory(tmpdir, httpbin):
    '''If we load a cassette in a nonexistent directory, it can save ok'''
    # Check to make sure directory doesnt exist
    assert not os.path.exists(str(tmpdir.join('nonexistent')))

    # Run VCR to create dir and cassette file
    with vcr.use_cassette(str(tmpdir.join('nonexistent', 'cassette.yml'))):
        urlopen(httpbin.url).read()

    # This should have made the file and the directory
    assert os.path.exists(str(tmpdir.join('nonexistent', 'cassette.yml')))


def test_unpatch(tmpdir, httpbin):
    '''Ensure that our cassette gets unpatched when we're done'''
    with vcr.use_cassette(str(tmpdir.join('unpatch.yaml'))) as cass:
        urlopen(httpbin.url).read()

    # Make the same request, and assert that we haven't served any more
    # requests out of cache
    urlopen(httpbin.url).read()
    assert cass.play_count == 0


def test_basic_json_use(tmpdir, httpbin):
    '''
    Ensure you can load a json serialized cassette
    '''
    test_fixture = str(tmpdir.join('synopsis.json'))
    with vcr.use_cassette(test_fixture, serializer='json'):
        response = urlopen(httpbin.url).read()
        assert b'difficult sometimes' in response


def test_patched_content(tmpdir, httpbin):
    '''
    Ensure that what you pull from a cassette is what came from the
    request
    '''
    with vcr.use_cassette(str(tmpdir.join('synopsis.yaml'))) as cass:
        response = urlopen(httpbin.url).read()
        assert cass.play_count == 0

    with vcr.use_cassette(str(tmpdir.join('synopsis.yaml'))) as cass:
        response2 = urlopen(httpbin.url).read()
        assert cass.play_count == 1
        cass._save(force=True)

    with vcr.use_cassette(str(tmpdir.join('synopsis.yaml'))) as cass:
        response3 = urlopen(httpbin.url).read()
        assert cass.play_count == 1

    assert response == response2
    assert response2 == response3


def test_patched_content_json(tmpdir, httpbin):
    '''
    Ensure that what you pull from a json cassette is what came from the
    request
    '''

    testfile = str(tmpdir.join('synopsis.json'))

    with vcr.use_cassette(testfile) as cass:
        response = urlopen(httpbin.url).read()
        assert cass.play_count == 0

    with vcr.use_cassette(testfile) as cass:
        response2 = urlopen(httpbin.url).read()
        assert cass.play_count == 1
        cass._save(force=True)

    with vcr.use_cassette(testfile) as cass:
        response3 = urlopen(httpbin.url).read()
        assert cass.play_count == 1

    assert response == response2
    assert response2 == response3



import os

import pytest
from six.moves import http_client as httplib

from vcr import VCR, use_cassette
from vcr.compat import mock
from vcr.request import Request
from vcr.stubs import VCRHTTPSConnection
from vcr.patch import _HTTPConnection, force_reset


def test_vcr_use_cassette():
    record_mode = mock.Mock()
    test_vcr = VCR(record_mode=record_mode)
    with mock.patch(
        'vcr.cassette.Cassette.load',
        return_value=mock.MagicMock(inject=False)
    ) as mock_cassette_load:

        @test_vcr.use_cassette('test')
        def function():
            pass

        assert mock_cassette_load.call_count == 0
        function()
        assert mock_cassette_load.call_args[1]['record_mode'] is record_mode

        # Make sure that calls to function now use cassettes with the
        # new filter_header_settings
        test_vcr.record_mode = mock.Mock()
        function()
        assert mock_cassette_load.call_args[1]['record_mode'] == test_vcr.record_mode

        # Ensure that explicitly provided arguments still supercede
        # those on the vcr.
        new_record_mode = mock.Mock()

    with test_vcr.use_cassette('test', record_mode=new_record_mode) as cassette:
        assert cassette.record_mode == new_record_mode


def test_vcr_before_record_request_params():
    base_path = 'http://httpbin.org/'

    def before_record_cb(request):
        if request.path != '/get':
            return request

    test_vcr = VCR(filter_headers=('cookie', ('bert', 'ernie')),
                   before_record_request=before_record_cb,
                   ignore_hosts=('www.test.com',), ignore_localhost=True,
                   filter_query_parameters=('foo', ('tom', 'jerry')),
                   filter_post_data_parameters=('posted', ('no', 'trespassing')))

    with test_vcr.use_cassette('test') as cassette:
        # Test explicit before_record_cb
        request_get = Request('GET', base_path + 'get', '', {})
        assert cassette.filter_request(request_get) is None
        request = Request('GET', base_path + 'get2', '', {})
        assert cassette.filter_request(request) is not None

        # Test filter_query_parameters
        request = Request('GET', base_path + '?foo=bar', '', {})
        assert cassette.filter_request(request).query == []
        request = Request('GET', base_path + '?tom=nobody', '', {})
        assert cassette.filter_request(request).query == [('tom', 'jerry')]

        # Test filter_headers
        request = Request('GET', base_path + '?foo=bar', '',
                          {'cookie': 'test', 'other': 'fun', 'bert': 'nobody'})
        assert (cassette.filter_request(request).headers ==
                {'other': 'fun', 'bert': 'ernie'})

        # Test ignore_hosts
        request = Request('GET', 'http://www.test.com' + '?foo=bar', '',
                          {'cookie': 'test', 'other': 'fun'})
        assert cassette.filter_request(request) is None

        # Test ignore_localhost
        request = Request('GET', 'http://localhost:8000' + '?foo=bar', '',
                          {'cookie': 'test', 'other': 'fun'})
        assert cassette.filter_request(request) is None

    with test_vcr.use_cassette('test', before_record_request=None) as cassette:
        # Test that before_record can be overwritten in context manager.
        assert cassette.filter_request(request_get) is not None


def test_vcr_before_record_response_iterable():
    # Regression test for #191

    request = Request('GET', '/', '', {})
    response = object()  # just can't be None

    # Prevent actually saving the cassette
    with mock.patch('vcr.cassette.save_cassette'):

        # Baseline: non-iterable before_record_response should work
        mock_filter = mock.Mock()
        vcr = VCR(before_record_response=mock_filter)
        with vcr.use_cassette('test') as cassette:
            assert mock_filter.call_count == 0
            cassette.append(request, response)
            assert mock_filter.call_count == 1

        # Regression test: iterable before_record_response should work too
        mock_filter = mock.Mock()
        vcr = VCR(before_record_response=(mock_filter,))
        with vcr.use_cassette('test') as cassette:
            assert mock_filter.call_count == 0
            cassette.append(request, response)
            assert mock_filter.call_count == 1


def test_before_record_response_as_filter():
    request = Request('GET', '/', '', {})
    response = object()  # just can't be None

    # Prevent actually saving the cassette
    with mock.patch('vcr.cassette.save_cassette'):

        filter_all = mock.Mock(return_value=None)
        vcr = VCR(before_record_response=filter_all)
        with vcr.use_cassette('test') as cassette:
            cassette.append(request, response)
            assert cassette.data == []
            assert not cassette.dirty


def test_vcr_path_transformer():
    # Regression test for #199

    # Prevent actually saving the cassette
    with mock.patch('vcr.cassette.save_cassette'):

        # Baseline: path should be unchanged
        vcr = VCR()
        with vcr.use_cassette('test') as cassette:
            assert cassette._path == 'test'

        # Regression test: path_transformer=None should do the same.
        vcr = VCR(path_transformer=None)
        with vcr.use_cassette('test') as cassette:
            assert cassette._path == 'test'

        # and it should still work with cassette_library_dir
        vcr = VCR(cassette_library_dir='/foo')
        with vcr.use_cassette('test') as cassette:
            assert cassette._path == '/foo/test'


@pytest.fixture
def random_fixture():
    return 1


@use_cassette('test')
def test_fixtures_with_use_cassette(random_fixture):
    # Applying a decorator to a test function that requests features can cause
    # problems if the decorator does not preserve the signature of the original
    # test function.

    # This test ensures that use_cassette preserves the signature of
    # the original test function, and thus that use_cassette is
    # compatible with py.test fixtures. It is admittedly a bit strange
    # because the test would never even run if the relevant feature
    # were broken.
    pass


def test_custom_patchers():
    class Test(object):
        attribute = None
        attribute2 = None
    test_vcr = VCR(custom_patches=((Test, 'attribute', VCRHTTPSConnection),))
    with test_vcr.use_cassette('custom_patches'):
        assert issubclass(Test.attribute, VCRHTTPSConnection)
        assert VCRHTTPSConnection is not Test.attribute

    with test_vcr.use_cassette(
        'custom_patches',
        custom_patches=((Test, 'attribute2', VCRHTTPSConnection),)
    ):
        assert issubclass(Test.attribute, VCRHTTPSConnection)
        assert VCRHTTPSConnection is not Test.attribute
        assert Test.attribute is Test.attribute2


def test_inject_cassette():
    vcr = VCR(inject_cassette=True)

    @vcr.use_cassette('test', record_mode='once')
    def with_cassette_injected(cassette):
        assert cassette.record_mode == 'once'

    @vcr.use_cassette('test', record_mode='once', inject_cassette=False)
    def without_cassette_injected():
        pass

    with_cassette_injected()
    without_cassette_injected()


def test_with_current_defaults():
    vcr = VCR(inject_cassette=True, record_mode='once')

    @vcr.use_cassette('test', with_current_defaults=False)
    def changing_defaults(cassette, checks):
        checks(cassette)

    @vcr.use_cassette('test', with_current_defaults=True)
    def current_defaults(cassette, checks):
        checks(cassette)

    def assert_record_mode_once(cassette):
        assert cassette.record_mode == 'once'

    def assert_record_mode_all(cassette):
        assert cassette.record_mode == 'all'

    changing_defaults(assert_record_mode_once)
    current_defaults(assert_record_mode_once)

    vcr.record_mode = 'all'
    changing_defaults(assert_record_mode_all)
    current_defaults(assert_record_mode_once)


def test_cassette_library_dir_with_decoration_and_no_explicit_path():
    library_dir = '/libary_dir'
    vcr = VCR(inject_cassette=True, cassette_library_dir=library_dir)

    @vcr.use_cassette()
    def function_name(cassette):
        assert cassette._path == os.path.join(library_dir, 'function_name')

    function_name()


def test_cassette_library_dir_with_decoration_and_explicit_path():
    library_dir = '/libary_dir'
    vcr = VCR(inject_cassette=True, cassette_library_dir=library_dir)

    @vcr.use_cassette(path='custom_name')
    def function_name(cassette):
        assert cassette._path == os.path.join(library_dir, 'custom_name')

    function_name()


def test_cassette_library_dir_with_decoration_and_super_explicit_path():
    library_dir = '/libary_dir'
    vcr = VCR(inject_cassette=True, cassette_library_dir=library_dir)

    @vcr.use_cassette(path=os.path.join(library_dir, 'custom_name'))
    def function_name(cassette):
        assert cassette._path == os.path.join(library_dir, 'custom_name')

    function_name()


def test_cassette_library_dir_with_path_transformer():
    library_dir = '/libary_dir'
    vcr = VCR(inject_cassette=True, cassette_library_dir=library_dir,
              path_transformer=lambda path: path + '.json')

    @vcr.use_cassette()
    def function_name(cassette):
        assert cassette._path == os.path.join(library_dir, 'function_name.json')

    function_name()


def test_use_cassette_with_no_extra_invocation():
    vcr = VCR(inject_cassette=True, cassette_library_dir='/')

    @vcr.use_cassette
    def function_name(cassette):
        assert cassette._path == os.path.join('/', 'function_name')

    function_name()


def test_path_transformer():
    vcr = VCR(inject_cassette=True, cassette_library_dir='/',
              path_transformer=lambda x: x + '_test')

    @vcr.use_cassette
    def function_name(cassette):
        assert cassette._path == os.path.join('/', 'function_name_test')

    function_name()


def test_cassette_name_generator_defaults_to_using_module_function_defined_in():
    vcr = VCR(inject_cassette=True)

    @vcr.use_cassette
    def function_name(cassette):
        assert cassette._path == os.path.join(os.path.dirname(__file__),
                                              'function_name')
    function_name()


def test_ensure_suffix():
    vcr = VCR(inject_cassette=True, path_transformer=VCR.ensure_suffix('.yaml'))

    @vcr.use_cassette
    def function_name(cassette):
        assert cassette._path == os.path.join(os.path.dirname(__file__),
                                              'function_name.yaml')

    function_name()


def test_additional_matchers():
    vcr = VCR(match_on=('uri',), inject_cassette=True)

    @vcr.use_cassette
    def function_defaults(cassette):
        assert set(cassette._match_on) == set([vcr.matchers['uri']])

    @vcr.use_cassette(additional_matchers=('body',))
    def function_additional(cassette):
        assert set(cassette._match_on) == set([vcr.matchers['uri'], vcr.matchers['body']])

    function_defaults()
    function_additional()


def test_decoration_should_respect_function_return_value():
    vcr = VCR()
    ret = 'a-return-value'

    @vcr.use_cassette
    def function_with_return():
        return ret

    assert ret == function_with_return()


class TestVCRClass(VCR().test_case()):

    def no_decoration(self):
        assert httplib.HTTPConnection == _HTTPConnection
        self.test_dynamically_added()
        assert httplib.HTTPConnection == _HTTPConnection

    def test_one(self):
        with force_reset():
            self.no_decoration()
        with force_reset():
            self.test_two()
        assert httplib.HTTPConnection != _HTTPConnection

    def test_two(self):
        assert httplib.HTTPConnection != _HTTPConnection


def test_dynamically_added(self):
    assert httplib.HTTPConnection != _HTTPConnection


TestVCRClass.test_dynamically_added = test_dynamically_added
del test_dynamically_added



import itertools

import pytest

from vcr import matchers
from vcr import request

# the dict contains requests with corresponding to its key difference
# with 'base' request.
REQUESTS = {
    'base': request.Request('GET', 'http://host.com/p?a=b', '', {}),
    'method': request.Request('POST', 'http://host.com/p?a=b', '', {}),
    'scheme': request.Request('GET', 'https://host.com:80/p?a=b', '', {}),
    'host': request.Request('GET', 'http://another-host.com/p?a=b', '', {}),
    'port': request.Request('GET', 'http://host.com:90/p?a=b', '', {}),
    'path': request.Request('GET', 'http://host.com/x?a=b', '', {}),
    'query': request.Request('GET', 'http://host.com/p?c=d', '', {}),
}


def assert_matcher(matcher_name):
    matcher = getattr(matchers, matcher_name)
    for k1, k2 in itertools.permutations(REQUESTS, 2):
        matched = matcher(REQUESTS[k1], REQUESTS[k2])
        if matcher_name in set((k1, k2)):
            assert not matched
        else:
            assert matched


def test_uri_matcher():
    for k1, k2 in itertools.permutations(REQUESTS, 2):
        matched = matchers.uri(REQUESTS[k1], REQUESTS[k2])
        if set((k1, k2)) != set(('base', 'method')):
            assert not matched
        else:
            assert matched


req1_body = (b"<?xml version='1.0'?><methodCall><methodName>test</methodName>"
             b"<params><param><value><array><data><value><struct>"
             b"<member><name>a</name><value><string>1</string></value></member>"
             b"<member><name>b</name><value><string>2</string></value></member>"
             b"</struct></value></data></array></value></param></params></methodCall>")
req2_body = (b"<?xml version='1.0'?><methodCall><methodName>test</methodName>"
             b"<params><param><value><array><data><value><struct>"
             b"<member><name>b</name><value><string>2</string></value></member>"
             b"<member><name>a</name><value><string>1</string></value></member>"
             b"</struct></value></data></array></value></param></params></methodCall>")


@pytest.mark.parametrize("r1, r2", [
    (
        request.Request('POST', 'http://host.com/', '123', {}),
        request.Request('POST', 'http://another-host.com/',
                        '123', {'Some-Header': 'value'})
    ),
    (
        request.Request('POST', 'http://host.com/', 'a=1&b=2',
                        {'Content-Type': 'application/x-www-form-urlencoded'}),
        request.Request('POST', 'http://host.com/', 'b=2&a=1',
                        {'Content-Type': 'application/x-www-form-urlencoded'})
    ),
    (
        request.Request('POST', 'http://host.com/', '123', {}),
        request.Request('POST', 'http://another-host.com/', '123', {'Some-Header': 'value'})
    ),
    (
        request.Request(
            'POST', 'http://host.com/', 'a=1&b=2',
            {'Content-Type': 'application/x-www-form-urlencoded'}
        ),
        request.Request(
            'POST', 'http://host.com/', 'b=2&a=1',
            {'Content-Type': 'application/x-www-form-urlencoded'}
        )
    ),
    (
        request.Request(
            'POST', 'http://host.com/', '{"a": 1, "b": 2}',
            {'Content-Type': 'application/json'}
        ),
        request.Request(
            'POST', 'http://host.com/', '{"b": 2, "a": 1}',
            {'content-type': 'application/json'}
        )
    ),
    (
        request.Request(
            'POST', 'http://host.com/', req1_body,
            {'User-Agent': 'xmlrpclib', 'Content-Type': 'text/xml'}
        ),
        request.Request(
            'POST', 'http://host.com/', req2_body,
            {'user-agent': 'somexmlrpc', 'content-type': 'text/xml'}
        )
    ),
    (
        request.Request(
            'POST', 'http://host.com/',
            '{"a": 1, "b": 2}', {'Content-Type': 'application/json'}
        ),
        request.Request(
            'POST', 'http://host.com/',
            '{"b": 2, "a": 1}', {'content-type': 'application/json'}
        )
    )
])
def test_body_matcher_does_match(r1, r2):
    assert matchers.body(r1, r2)


@pytest.mark.parametrize("r1, r2", [
    (
        request.Request('POST', 'http://host.com/', '{"a": 1, "b": 2}', {}),
        request.Request('POST', 'http://host.com/', '{"b": 2, "a": 1}', {}),
    ),
    (
        request.Request(
            'POST', 'http://host.com/',
            '{"a": 1, "b": 3}', {'Content-Type': 'application/json'}
        ),
        request.Request(
            'POST', 'http://host.com/',
            '{"b": 2, "a": 1}', {'content-type': 'application/json'}
        )
    ),
    (
        request.Request(
            'POST', 'http://host.com/', req1_body, {'Content-Type': 'text/xml'}
        ),
        request.Request(
            'POST', 'http://host.com/', req2_body, {'content-type': 'text/xml'}
        )
    )
])
def test_body_match_does_not_match(r1, r2):
    assert not matchers.body(r1, r2)


def test_query_matcher():
    req1 = request.Request('GET', 'http://host.com/?a=b&c=d', '', {})
    req2 = request.Request('GET', 'http://host.com/?c=d&a=b', '', {})
    assert matchers.query(req1, req2)

    req1 = request.Request('GET', 'http://host.com/?a=b&a=b&c=d', '', {})
    req2 = request.Request('GET', 'http://host.com/?a=b&c=d&a=b', '', {})
    req3 = request.Request('GET', 'http://host.com/?c=d&a=b&a=b', '', {})
    assert matchers.query(req1, req2)
    assert matchers.query(req1, req3)


def test_metchers():
    assert_matcher('method')
    assert_matcher('scheme')
    assert_matcher('host')
    assert_matcher('port')
    assert_matcher('path')
    assert_matcher('query')



# -*- coding: utf-8 -*-
'''Test requests' interaction with vcr'''

import json

import pytest
import vcr
from vcr.errors import CannotOverwriteExistingCassetteException

from assertions import assert_cassette_empty, assert_is_json

tornado = pytest.importorskip("tornado")
http = pytest.importorskip("tornado.httpclient")

# whether the current version of Tornado supports the raise_error argument for
# fetch().
supports_raise_error = tornado.version_info >= (4,)


@pytest.fixture(params=['simple', 'curl', 'default'])
def get_client(request):
    if request.param == 'simple':
        from tornado import simple_httpclient as simple
        return (lambda: simple.SimpleAsyncHTTPClient())
    elif request.param == 'curl':
        curl = pytest.importorskip("tornado.curl_httpclient")
        return (lambda: curl.CurlAsyncHTTPClient())
    else:
        return (lambda: http.AsyncHTTPClient())


def get(client, url, **kwargs):
    fetch_kwargs = {}
    if supports_raise_error:
        fetch_kwargs['raise_error'] = kwargs.pop('raise_error', True)

    return client.fetch(
        http.HTTPRequest(url, method='GET', **kwargs),
        **fetch_kwargs
    )


def post(client, url, data=None, **kwargs):
    if data:
        kwargs['body'] = json.dumps(data)
    return client.fetch(http.HTTPRequest(url, method='POST', **kwargs))


@pytest.fixture(params=["https", "http"])
def scheme(request):
    '''Fixture that returns both http and https.'''
    return request.param


@pytest.mark.gen_test
def test_status_code(get_client, scheme, tmpdir):
    '''Ensure that we can read the status code'''
    url = scheme + '://httpbin.org/'
    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        status_code = (yield get(get_client(), url)).code

    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))) as cass:
        assert status_code == (yield get(get_client(), url)).code
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_headers(get_client, scheme, tmpdir):
    '''Ensure that we can read the headers back'''
    url = scheme + '://httpbin.org/'
    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        headers = (yield get(get_client(), url)).headers

    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))) as cass:
        assert headers == (yield get(get_client(), url)).headers
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_body(get_client, tmpdir, scheme):
    '''Ensure the responses are all identical enough'''

    url = scheme + '://httpbin.org/bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        content = (yield get(get_client(), url)).body

    with vcr.use_cassette(str(tmpdir.join('body.yaml'))) as cass:
        assert content == (yield get(get_client(), url)).body
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_effective_url(get_client, scheme, tmpdir):
    '''Ensure that the effective_url is captured'''
    url = scheme + '://httpbin.org/redirect-to?url=/html'
    with vcr.use_cassette(str(tmpdir.join('url.yaml'))):
        effective_url = (yield get(get_client(), url)).effective_url
        assert effective_url == scheme + '://httpbin.org/html'

    with vcr.use_cassette(str(tmpdir.join('url.yaml'))) as cass:
        assert effective_url == (yield get(get_client(), url)).effective_url
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_auth(get_client, tmpdir, scheme):
    '''Ensure that we can handle basic auth'''
    auth = ('user', 'passwd')
    url = scheme + '://httpbin.org/basic-auth/user/passwd'
    with vcr.use_cassette(str(tmpdir.join('auth.yaml'))):
        one = yield get(
            get_client(), url, auth_username=auth[0], auth_password=auth[1]
        )

    with vcr.use_cassette(str(tmpdir.join('auth.yaml'))) as cass:
        two = yield get(
            get_client(), url, auth_username=auth[0], auth_password=auth[1]
        )
        assert one.body == two.body
        assert one.code == two.code
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_auth_failed(get_client, tmpdir, scheme):
    '''Ensure that we can save failed auth statuses'''
    auth = ('user', 'wrongwrongwrong')
    url = scheme + '://httpbin.org/basic-auth/user/passwd'
    with vcr.use_cassette(str(tmpdir.join('auth-failed.yaml'))) as cass:
        # Ensure that this is empty to begin with
        assert_cassette_empty(cass)
        with pytest.raises(http.HTTPError) as exc_info:
            yield get(
                get_client(),
                url,
                auth_username=auth[0],
                auth_password=auth[1],
            )
        one = exc_info.value.response
        assert exc_info.value.code == 401

    with vcr.use_cassette(str(tmpdir.join('auth-failed.yaml'))) as cass:
        with pytest.raises(http.HTTPError) as exc_info:
            two = yield get(
                get_client(),
                url,
                auth_username=auth[0],
                auth_password=auth[1],
            )
        two = exc_info.value.response
        assert exc_info.value.code == 401
        assert one.body == two.body
        assert one.code == two.code == 401
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_post(get_client, tmpdir, scheme):
    '''Ensure that we can post and cache the results'''
    data = {'key1': 'value1', 'key2': 'value2'}
    url = scheme + '://httpbin.org/post'
    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        req1 = (yield post(get_client(), url, data)).body

    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))) as cass:
        req2 = (yield post(get_client(), url, data)).body

    assert req1 == req2
    assert 1 == cass.play_count


@pytest.mark.gen_test
def test_redirects(get_client, tmpdir, scheme):
    '''Ensure that we can handle redirects'''
    url = scheme + '://httpbin.org/redirect-to?url=bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        content = (yield get(get_client(), url)).body

    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))) as cass:
        assert content == (yield get(get_client(), url)).body
        assert cass.play_count == 1


@pytest.mark.gen_test
def test_cross_scheme(get_client, tmpdir, scheme):
    '''Ensure that requests between schemes are treated separately'''
    # First fetch a url under http, and then again under https and then
    # ensure that we haven't served anything out of cache, and we have two
    # requests / response pairs in the cassette
    with vcr.use_cassette(str(tmpdir.join('cross_scheme.yaml'))) as cass:
        yield get(get_client(), 'https://httpbin.org/')
        yield get(get_client(), 'http://httpbin.org/')
        assert cass.play_count == 0
        assert len(cass) == 2

    # Then repeat the same requests and ensure both were replayed.
    with vcr.use_cassette(str(tmpdir.join('cross_scheme.yaml'))) as cass:
        yield get(get_client(), 'https://httpbin.org/')
        yield get(get_client(), 'http://httpbin.org/')
        assert cass.play_count == 2


@pytest.mark.gen_test
def test_gzip(get_client, tmpdir, scheme):
    '''
    Ensure that httpclient is able to automatically decompress the response
    body
    '''
    url = scheme + '://httpbin.org/gzip'

    # use_gzip was renamed to decompress_response in 4.0
    kwargs = {}
    if tornado.version_info < (4,):
        kwargs['use_gzip'] = True
    else:
        kwargs['decompress_response'] = True

    with vcr.use_cassette(str(tmpdir.join('gzip.yaml'))):
        response = yield get(get_client(), url, **kwargs)
        assert_is_json(response.body)

    with vcr.use_cassette(str(tmpdir.join('gzip.yaml'))) as cass:
        response = yield get(get_client(), url, **kwargs)
        assert_is_json(response.body)
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_https_with_cert_validation_disabled(get_client, tmpdir):
    cass_path = str(tmpdir.join('cert_validation_disabled.yaml'))

    with vcr.use_cassette(cass_path):
        yield get(get_client(), 'https://httpbin.org', validate_cert=False)

    with vcr.use_cassette(cass_path) as cass:
        yield get(get_client(), 'https://httpbin.org', validate_cert=False)
        assert 1 == cass.play_count


@pytest.mark.gen_test
def test_unsupported_features_raises_in_future(get_client, tmpdir):
    '''Ensure that the exception for an AsyncHTTPClient feature not being
    supported is raised inside the future.'''

    def callback(chunk):
        assert False, "Did not expect to be called."

    with vcr.use_cassette(str(tmpdir.join('invalid.yaml'))):
        future = get(
            get_client(), 'http://httpbin.org', streaming_callback=callback
        )

    with pytest.raises(Exception) as excinfo:
        yield future

    assert "not yet supported by VCR" in str(excinfo)


@pytest.mark.skipif(
    not supports_raise_error,
    reason='raise_error unavailable in tornado <= 3',
)
@pytest.mark.gen_test
def test_unsupported_features_raise_error_disabled(get_client, tmpdir):
    '''Ensure that the exception for an AsyncHTTPClient feature not being
    supported is not raised if raise_error=False.'''

    def callback(chunk):
        assert False, "Did not expect to be called."

    with vcr.use_cassette(str(tmpdir.join('invalid.yaml'))):
        response = yield get(
            get_client(),
            'http://httpbin.org',
            streaming_callback=callback,
            raise_error=False,
        )

    assert "not yet supported by VCR" in str(response.error)


@pytest.mark.gen_test
def test_cannot_overwrite_cassette_raises_in_future(get_client, tmpdir):
    '''Ensure that CannotOverwriteExistingCassetteException is raised inside
    the future.'''

    with vcr.use_cassette(str(tmpdir.join('overwrite.yaml'))):
        yield get(get_client(), 'http://httpbin.org/get')

    with vcr.use_cassette(str(tmpdir.join('overwrite.yaml'))):
        future = get(get_client(), 'http://httpbin.org/headers')

    with pytest.raises(CannotOverwriteExistingCassetteException):
        yield future


@pytest.mark.skipif(
    not supports_raise_error,
    reason='raise_error unavailable in tornado <= 3',
)
@pytest.mark.gen_test
def test_cannot_overwrite_cassette_raise_error_disabled(get_client, tmpdir):
    '''Ensure that CannotOverwriteExistingCassetteException is not raised if
    raise_error=False in the fetch() call.'''

    with vcr.use_cassette(str(tmpdir.join('overwrite.yaml'))):
        yield get(
            get_client(), 'http://httpbin.org/get', raise_error=False
        )

    with vcr.use_cassette(str(tmpdir.join('overwrite.yaml'))):
        response = yield get(
            get_client(), 'http://httpbin.org/headers', raise_error=False
        )

    assert isinstance(response.error, CannotOverwriteExistingCassetteException)


@pytest.mark.gen_test
@vcr.use_cassette(path_transformer=vcr.default_vcr.ensure_suffix('.yaml'))
def test_tornado_with_decorator_use_cassette(get_client):
    response = yield get_client().fetch(
        http.HTTPRequest('http://www.google.com/', method='GET')
    )
    assert response.body.decode('utf-8') == "not actually google"


@pytest.mark.gen_test
@vcr.use_cassette(path_transformer=vcr.default_vcr.ensure_suffix('.yaml'))
def test_tornado_exception_can_be_caught(get_client):
    try:
        yield get(get_client(), 'http://httpbin.org/status/500')
    except http.HTTPError as e:
        assert e.code == 500

    try:
        yield get(get_client(), 'http://httpbin.org/status/404')
    except http.HTTPError as e:
        assert e.code == 404


@pytest.mark.gen_test
def test_existing_references_get_patched(tmpdir):
    from tornado.httpclient import AsyncHTTPClient

    with vcr.use_cassette(str(tmpdir.join('data.yaml'))):
        client = AsyncHTTPClient()
        yield get(client, 'http://httpbin.org/get')

    with vcr.use_cassette(str(tmpdir.join('data.yaml'))) as cass:
        yield get(client, 'http://httpbin.org/get')
        assert cass.play_count == 1


@pytest.mark.gen_test
def test_existing_instances_get_patched(get_client, tmpdir):
    '''Ensure that existing instances of AsyncHTTPClient get patched upon
    entering VCR context.'''

    client = get_client()

    with vcr.use_cassette(str(tmpdir.join('data.yaml'))):
        yield get(client, 'http://httpbin.org/get')

    with vcr.use_cassette(str(tmpdir.join('data.yaml'))) as cass:
        yield get(client, 'http://httpbin.org/get')
        assert cass.play_count == 1


@pytest.mark.gen_test
def test_request_time_is_set(get_client, tmpdir):
    '''Ensures that the request_time on HTTPResponses is set.'''

    with vcr.use_cassette(str(tmpdir.join('data.yaml'))):
        client = get_client()
        response = yield get(client, 'http://httpbin.org/get')
        assert response.request_time is not None

    with vcr.use_cassette(str(tmpdir.join('data.yaml'))) as cass:
        client = get_client()
        response = yield get(client, 'http://httpbin.org/get')
        assert response.request_time is not None
        assert cass.play_count == 1



'''Utilities for patching in cassettes'''
import functools
import itertools

from .compat import contextlib, mock
from .stubs import VCRHTTPConnection, VCRHTTPSConnection
from six.moves import http_client as httplib


# Save some of the original types for the purposes of unpatching
_HTTPConnection = httplib.HTTPConnection
_HTTPSConnection = httplib.HTTPSConnection


# Try to save the original types for requests
try:
    import requests.packages.urllib3.connectionpool as cpool
except ImportError:  # pragma: no cover
    pass
else:
    _VerifiedHTTPSConnection = cpool.VerifiedHTTPSConnection
    _cpoolHTTPConnection = cpool.HTTPConnection
    _cpoolHTTPSConnection = cpool.HTTPSConnection

# Try to save the original types for boto3
try:
    import botocore.vendored.requests.packages.urllib3.connectionpool as cpool
except ImportError:  # pragma: no cover
    pass
else:
    _Boto3VerifiedHTTPSConnection = cpool.VerifiedHTTPSConnection
    _cpoolBoto3HTTPConnection = cpool.HTTPConnection
    _cpoolBoto3HTTPSConnection = cpool.HTTPSConnection


# Try to save the original types for urllib3
try:
    import urllib3
except ImportError:  # pragma: no cover
    pass
else:
    _VerifiedHTTPSConnection = urllib3.connectionpool.VerifiedHTTPSConnection


# Try to save the original types for httplib2
try:
    import httplib2
except ImportError:  # pragma: no cover
    pass
else:
    _HTTPConnectionWithTimeout = httplib2.HTTPConnectionWithTimeout
    _HTTPSConnectionWithTimeout = httplib2.HTTPSConnectionWithTimeout
    _SCHEME_TO_CONNECTION = httplib2.SCHEME_TO_CONNECTION


# Try to save the original types for boto
try:
    import boto.https_connection
except ImportError:  # pragma: no cover
    pass
else:
    _CertValidatingHTTPSConnection = boto.https_connection.CertValidatingHTTPSConnection


# Try to save the original types for Tornado
try:
    import tornado.simple_httpclient
except ImportError:  # pragma: no cover
    pass
else:
    _SimpleAsyncHTTPClient_fetch_impl = \
        tornado.simple_httpclient.SimpleAsyncHTTPClient.fetch_impl


try:
    import tornado.curl_httpclient
except ImportError:  # pragma: no cover
    pass
else:
    _CurlAsyncHTTPClient_fetch_impl = \
        tornado.curl_httpclient.CurlAsyncHTTPClient.fetch_impl

try:
    import aiohttp.client
except ImportError:  # pragma: no cover
    pass
else:
    _AiohttpClientSessionRequest = aiohttp.client.ClientSession._request


class CassettePatcherBuilder(object):

    def _build_patchers_from_mock_triples_decorator(function):
        @functools.wraps(function)
        def wrapped(self, *args, **kwargs):
            return self._build_patchers_from_mock_triples(
                function(self, *args, **kwargs)
            )
        return wrapped

    def __init__(self, cassette):
        self._cassette = cassette
        self._class_to_cassette_subclass = {}

    def build(self):
        return itertools.chain(
            self._httplib(), self._requests(), self._boto3(), self._urllib3(),
            self._httplib2(), self._boto(), self._tornado(), self._aiohttp(),
            self._build_patchers_from_mock_triples(
                self._cassette.custom_patches
            ),
        )

    def _build_patchers_from_mock_triples(self, mock_triples):
        for args in mock_triples:
            patcher = self._build_patcher(*args)
            if patcher:
                yield patcher

    def _build_patcher(self, obj, patched_attribute, replacement_class):
        if not hasattr(obj, patched_attribute):
            return

        return mock.patch.object(obj, patched_attribute,
                                 self._recursively_apply_get_cassette_subclass(
                                     replacement_class))

    def _recursively_apply_get_cassette_subclass(self, replacement_dict_or_obj):
        """One of the subtleties of this class is that it does not directly
        replace HTTPSConnection with `VCRRequestsHTTPSConnection`, but a
        subclass of the aforementioned class that has the `cassette`
        class attribute assigned to `self._cassette`. This behavior is
        necessary to properly support nested cassette contexts.

        This function exists to ensure that we use the same class
        object (reference) to patch everything that replaces
        VCRRequestHTTP[S]Connection, but that we can talk about
        patching them with the raw references instead, and without
        worrying about exactly where the subclass with the relevant
        value for `cassette` is first created.

        The function is recursive because it looks in to dictionaries
        and replaces class values at any depth with the subclass
        described in the previous paragraph.
        """
        if isinstance(replacement_dict_or_obj, dict):
            for key, replacement_obj in replacement_dict_or_obj.items():
                replacement_obj = self._recursively_apply_get_cassette_subclass(
                    replacement_obj)
                replacement_dict_or_obj[key] = replacement_obj
            return replacement_dict_or_obj
        if hasattr(replacement_dict_or_obj, 'cassette'):
            replacement_dict_or_obj = self._get_cassette_subclass(
                replacement_dict_or_obj)
        return replacement_dict_or_obj

    def _get_cassette_subclass(self, klass):
        if klass.cassette is not None:
            return klass
        if klass not in self._class_to_cassette_subclass:
            subclass = self._build_cassette_subclass(klass)
            self._class_to_cassette_subclass[klass] = subclass
        return self._class_to_cassette_subclass[klass]

    def _build_cassette_subclass(self, base_class):
        bases = (base_class,)
        if not issubclass(base_class, object):  # Check for old style class
            bases += (object,)
        return type('{0}{1}'.format(base_class.__name__, self._cassette._path),
                    bases, dict(cassette=self._cassette))

    @_build_patchers_from_mock_triples_decorator
    def _httplib(self):
        yield httplib, 'HTTPConnection', VCRHTTPConnection
        yield httplib, 'HTTPSConnection', VCRHTTPSConnection

    def _requests(self):
        try:
            import requests.packages.urllib3.connectionpool as cpool
        except ImportError:  # pragma: no cover
            return ()
        from .stubs import requests_stubs
        return self._urllib3_patchers(cpool, requests_stubs)

    def _boto3(self):
        try:
            import botocore.vendored.requests.packages.urllib3.connectionpool as cpool
        except ImportError:  # pragma: no cover
            return ()
        from .stubs import boto3_stubs
        return self._urllib3_patchers(cpool, boto3_stubs)

    def _patched_get_conn(self, connection_pool_class, connection_class_getter):
        get_conn = connection_pool_class._get_conn

        @functools.wraps(get_conn)
        def patched_get_conn(pool, timeout=None):
            connection = get_conn(pool, timeout)
            connection_class = (
                pool.ConnectionCls if hasattr(pool, 'ConnectionCls')
                else connection_class_getter())
            # We need to make sure that we are actually providing a
            # patched version of the connection class. This might not
            # always be the case because the pool keeps previously
            # used connections (which might actually be of a different
            # class) around. This while loop will terminate because
            # eventually the pool will run out of connections.
            while not isinstance(connection, connection_class):
                connection = get_conn(pool, timeout)
            return connection

        return patched_get_conn

    def _patched_new_conn(self, connection_pool_class, connection_remover):
        new_conn = connection_pool_class._new_conn

        @functools.wraps(new_conn)
        def patched_new_conn(pool):
            new_connection = new_conn(pool)
            connection_remover.add_connection_to_pool_entry(pool, new_connection)
            return new_connection

        return patched_new_conn

    def _urllib3(self):
        try:
            import urllib3.connectionpool as cpool
        except ImportError:  # pragma: no cover
            return ()
        from .stubs import urllib3_stubs
        return self._urllib3_patchers(cpool, urllib3_stubs)

    @_build_patchers_from_mock_triples_decorator
    def _httplib2(self):
        try:
            import httplib2 as cpool
        except ImportError:  # pragma: no cover
            pass
        else:
            from .stubs.httplib2_stubs import VCRHTTPConnectionWithTimeout
            from .stubs.httplib2_stubs import VCRHTTPSConnectionWithTimeout

            yield cpool, 'HTTPConnectionWithTimeout', VCRHTTPConnectionWithTimeout
            yield cpool, 'HTTPSConnectionWithTimeout', VCRHTTPSConnectionWithTimeout
            yield cpool, 'SCHEME_TO_CONNECTION', {'http': VCRHTTPConnectionWithTimeout,
                                                  'https': VCRHTTPSConnectionWithTimeout}

    @_build_patchers_from_mock_triples_decorator
    def _boto(self):
        try:
            import boto.https_connection as cpool
        except ImportError:  # pragma: no cover
            pass
        else:
            from .stubs.boto_stubs import VCRCertValidatingHTTPSConnection
            yield cpool, 'CertValidatingHTTPSConnection', VCRCertValidatingHTTPSConnection

    @_build_patchers_from_mock_triples_decorator
    def _tornado(self):
        try:
            import tornado.simple_httpclient as simple
        except ImportError:  # pragma: no cover
            pass
        else:
            from .stubs.tornado_stubs import vcr_fetch_impl

            new_fetch_impl = vcr_fetch_impl(
                self._cassette, _SimpleAsyncHTTPClient_fetch_impl
            )
            yield simple.SimpleAsyncHTTPClient, 'fetch_impl', new_fetch_impl
        try:
            import tornado.curl_httpclient as curl
        except ImportError:  # pragma: no cover
            pass
        else:
            from .stubs.tornado_stubs import vcr_fetch_impl

            new_fetch_impl = vcr_fetch_impl(
                self._cassette, _CurlAsyncHTTPClient_fetch_impl
            )
            yield curl.CurlAsyncHTTPClient, 'fetch_impl', new_fetch_impl

    @_build_patchers_from_mock_triples_decorator
    def _aiohttp(self):
        try:
            import aiohttp.client as client
        except ImportError:  # pragma: no cover
            pass
        else:
            from .stubs.aiohttp_stubs import vcr_request
            new_request = vcr_request(
                self._cassette, _AiohttpClientSessionRequest
            )
            yield client.ClientSession, '_request', new_request

    def _urllib3_patchers(self, cpool, stubs):
        http_connection_remover = ConnectionRemover(
            self._get_cassette_subclass(stubs.VCRRequestsHTTPConnection)
        )
        https_connection_remover = ConnectionRemover(
            self._get_cassette_subclass(stubs.VCRRequestsHTTPSConnection)
        )
        mock_triples = (
            (cpool, 'VerifiedHTTPSConnection', stubs.VCRRequestsHTTPSConnection),
            (cpool, 'VerifiedHTTPSConnection', stubs.VCRRequestsHTTPSConnection),
            (cpool, 'HTTPConnection', stubs.VCRRequestsHTTPConnection),
            (cpool, 'HTTPSConnection', stubs.VCRRequestsHTTPSConnection),
            (cpool, 'is_connection_dropped', mock.Mock(return_value=False)),  # Needed on Windows only
            (cpool.HTTPConnectionPool, 'ConnectionCls', stubs.VCRRequestsHTTPConnection),
            (cpool.HTTPSConnectionPool, 'ConnectionCls', stubs.VCRRequestsHTTPSConnection),
        )
        # These handle making sure that sessions only use the
        # connections of the appropriate type.
        mock_triples += ((cpool.HTTPConnectionPool, '_get_conn',
                          self._patched_get_conn(cpool.HTTPConnectionPool,
                                                 lambda: cpool.HTTPConnection)),
                         (cpool.HTTPSConnectionPool, '_get_conn',
                          self._patched_get_conn(cpool.HTTPSConnectionPool,
                                                 lambda: cpool.HTTPSConnection)),
                         (cpool.HTTPConnectionPool, '_new_conn',
                          self._patched_new_conn(cpool.HTTPConnectionPool,
                                                 http_connection_remover)),
                         (cpool.HTTPSConnectionPool, '_new_conn',
                          self._patched_new_conn(cpool.HTTPSConnectionPool,
                                                 https_connection_remover)))

        return itertools.chain(self._build_patchers_from_mock_triples(mock_triples),
                               (http_connection_remover, https_connection_remover))


class ConnectionRemover(object):

    def __init__(self, connection_class):
        self._connection_class = connection_class
        self._connection_pool_to_connections = {}

    def add_connection_to_pool_entry(self, pool, connection):
        if isinstance(connection, self._connection_class):
            self._connection_pool_to_connections.setdefault(pool, set()).add(connection)

    def remove_connection_to_pool_entry(self, pool, connection):
        if isinstance(connection, self._connection_class):
            self._connection_pool_to_connections[self._connection_class].remove(connection)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        for pool, connections in self._connection_pool_to_connections.items():
            readd_connections = []
            while pool.pool and not pool.pool.empty() and connections:
                connection = pool.pool.get()
                if isinstance(connection, self._connection_class):
                    connections.remove(connection)
                else:
                    readd_connections.append(connection)
            for connection in readd_connections:
                pool._put_conn(connection)


def reset_patchers():
    yield mock.patch.object(httplib, 'HTTPConnection', _HTTPConnection)
    yield mock.patch.object(httplib, 'HTTPSConnection', _HTTPSConnection)
    try:
        import requests.packages.urllib3.connectionpool as cpool
    except ImportError:  # pragma: no cover
        pass
    else:
        # unpatch requests v1.x
        yield mock.patch.object(cpool, 'VerifiedHTTPSConnection', _VerifiedHTTPSConnection)
        yield mock.patch.object(cpool, 'HTTPConnection', _cpoolHTTPConnection)
        # unpatch requests v2.x
        if hasattr(cpool.HTTPConnectionPool, 'ConnectionCls'):
            yield mock.patch.object(cpool.HTTPConnectionPool, 'ConnectionCls',
                                    _cpoolHTTPConnection)
            yield mock.patch.object(cpool.HTTPSConnectionPool, 'ConnectionCls',
                                    _cpoolHTTPSConnection)

        if hasattr(cpool, 'HTTPSConnection'):
            yield mock.patch.object(cpool, 'HTTPSConnection', _cpoolHTTPSConnection)

    try:
        import urllib3.connectionpool as cpool
    except ImportError:  # pragma: no cover
        pass
    else:
        yield mock.patch.object(cpool, 'VerifiedHTTPSConnection', _VerifiedHTTPSConnection)
        yield mock.patch.object(cpool, 'HTTPConnection', _HTTPConnection)
        yield mock.patch.object(cpool, 'HTTPSConnection', _HTTPSConnection)
        if hasattr(cpool.HTTPConnectionPool, 'ConnectionCls'):
            yield mock.patch.object(cpool.HTTPConnectionPool, 'ConnectionCls', _HTTPConnection)
            yield mock.patch.object(cpool.HTTPSConnectionPool, 'ConnectionCls', _HTTPSConnection)

    try:
        import botocore.vendored.requests.packages.urllib3.connectionpool as cpool
    except ImportError:  # pragma: no cover
        pass
    else:
        # unpatch requests v1.x
        yield mock.patch.object(cpool, 'VerifiedHTTPSConnection', _Boto3VerifiedHTTPSConnection)
        yield mock.patch.object(cpool, 'HTTPConnection', _cpoolBoto3HTTPConnection)
        # unpatch requests v2.x
        if hasattr(cpool.HTTPConnectionPool, 'ConnectionCls'):
            yield mock.patch.object(cpool.HTTPConnectionPool, 'ConnectionCls',
                                    _cpoolBoto3HTTPConnection)
            yield mock.patch.object(cpool.HTTPSConnectionPool, 'ConnectionCls',
                                    _cpoolBoto3HTTPSConnection)

        if hasattr(cpool, 'HTTPSConnection'):
            yield mock.patch.object(cpool, 'HTTPSConnection', _cpoolBoto3HTTPSConnection)

    try:
        import httplib2 as cpool
    except ImportError:  # pragma: no cover
        pass
    else:
        yield mock.patch.object(cpool, 'HTTPConnectionWithTimeout', _HTTPConnectionWithTimeout)
        yield mock.patch.object(cpool, 'HTTPSConnectionWithTimeout', _HTTPSConnectionWithTimeout)
        yield mock.patch.object(cpool, 'SCHEME_TO_CONNECTION', _SCHEME_TO_CONNECTION)

    try:
        import boto.https_connection as cpool
    except ImportError:  # pragma: no cover
        pass
    else:
        yield mock.patch.object(cpool, 'CertValidatingHTTPSConnection',
                                _CertValidatingHTTPSConnection)

    try:
        import tornado.simple_httpclient as simple
    except ImportError:  # pragma: no cover
        pass
    else:
        yield mock.patch.object(
            simple.SimpleAsyncHTTPClient,
            'fetch_impl',
            _SimpleAsyncHTTPClient_fetch_impl,
        )
    try:
        import tornado.curl_httpclient as curl
    except ImportError:  # pragma: no cover
        pass
    else:
        yield mock.patch.object(
            curl.CurlAsyncHTTPClient,
            'fetch_impl',
            _CurlAsyncHTTPClient_fetch_impl,
        )


@contextlib.contextmanager
def force_reset():
    with contextlib.ExitStack() as exit_stack:
        for patcher in reset_patchers():
            exit_stack.enter_context(patcher)
        yield



'''Integration tests with urllib3'''

# coding=utf-8

import pytest
import pytest_httpbin
import vcr
from assertions import assert_cassette_empty, assert_is_json
urllib3 = pytest.importorskip("urllib3")


@pytest.fixture(scope='module')
def verify_pool_mgr():
    return urllib3.PoolManager(
        cert_reqs='CERT_REQUIRED',  # Force certificate check.
        ca_certs=pytest_httpbin.certs.where()
    )


@pytest.fixture(scope='module')
def pool_mgr():
    return urllib3.PoolManager()


def test_status_code(httpbin_both, tmpdir, verify_pool_mgr):
    '''Ensure that we can read the status code'''
    url = httpbin_both.url
    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        status_code = verify_pool_mgr.request('GET', url).status

    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        assert status_code == verify_pool_mgr.request('GET', url).status


def test_headers(tmpdir, httpbin_both, verify_pool_mgr):
    '''Ensure that we can read the headers back'''
    url = httpbin_both.url
    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        headers = verify_pool_mgr.request('GET', url).headers

    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        assert headers == verify_pool_mgr.request('GET', url).headers


def test_body(tmpdir, httpbin_both, verify_pool_mgr):
    '''Ensure the responses are all identical enough'''
    url = httpbin_both.url + '/bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        content = verify_pool_mgr.request('GET', url).data

    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        assert content == verify_pool_mgr.request('GET', url).data


def test_auth(tmpdir, httpbin_both, verify_pool_mgr):
    '''Ensure that we can handle basic auth'''
    auth = ('user', 'passwd')
    headers = urllib3.util.make_headers(basic_auth='{0}:{1}'.format(*auth))
    url = httpbin_both.url + '/basic-auth/user/passwd'
    with vcr.use_cassette(str(tmpdir.join('auth.yaml'))):
        one = verify_pool_mgr.request('GET', url, headers=headers)

    with vcr.use_cassette(str(tmpdir.join('auth.yaml'))):
        two = verify_pool_mgr.request('GET', url, headers=headers)
        assert one.data == two.data
        assert one.status == two.status


def test_auth_failed(tmpdir, httpbin_both, verify_pool_mgr):
    '''Ensure that we can save failed auth statuses'''
    auth = ('user', 'wrongwrongwrong')
    headers = urllib3.util.make_headers(basic_auth='{0}:{1}'.format(*auth))
    url = httpbin_both.url + '/basic-auth/user/passwd'
    with vcr.use_cassette(str(tmpdir.join('auth-failed.yaml'))) as cass:
        # Ensure that this is empty to begin with
        assert_cassette_empty(cass)
        one = verify_pool_mgr.request('GET', url, headers=headers)
        two = verify_pool_mgr.request('GET', url, headers=headers)
        assert one.data == two.data
        assert one.status == two.status == 401


def test_post(tmpdir, httpbin_both, verify_pool_mgr):
    '''Ensure that we can post and cache the results'''
    data = {'key1': 'value1', 'key2': 'value2'}
    url = httpbin_both.url + '/post'
    with vcr.use_cassette(str(tmpdir.join('verify_pool_mgr.yaml'))):
        req1 = verify_pool_mgr.request('POST', url, data).data

    with vcr.use_cassette(str(tmpdir.join('verify_pool_mgr.yaml'))):
        req2 = verify_pool_mgr.request('POST', url, data).data

    assert req1 == req2


def test_redirects(tmpdir, httpbin_both, verify_pool_mgr):
    '''Ensure that we can handle redirects'''
    url = httpbin_both.url + '/redirect-to?url=bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('verify_pool_mgr.yaml'))):
        content = verify_pool_mgr.request('GET', url).data

    with vcr.use_cassette(str(tmpdir.join('verify_pool_mgr.yaml'))) as cass:
        assert content == verify_pool_mgr.request('GET', url).data
        # Ensure that we've now cached *two* responses. One for the redirect
        # and one for the final fetch
        assert len(cass) == 2
        assert cass.play_count == 2


def test_cross_scheme(tmpdir, httpbin, httpbin_secure, verify_pool_mgr):
    '''Ensure that requests between schemes are treated separately'''
    # First fetch a url under http, and then again under https and then
    # ensure that we haven't served anything out of cache, and we have two
    # requests / response pairs in the cassette
    with vcr.use_cassette(str(tmpdir.join('cross_scheme.yaml'))) as cass:
        verify_pool_mgr.request('GET', httpbin_secure.url)
        verify_pool_mgr.request('GET', httpbin.url)
        assert cass.play_count == 0
        assert len(cass) == 2


def test_gzip(tmpdir, httpbin_both, verify_pool_mgr):
    '''
    Ensure that requests (actually urllib3) is able to automatically decompress
    the response body
    '''
    url = httpbin_both.url + '/gzip'
    response = verify_pool_mgr.request('GET', url)

    with vcr.use_cassette(str(tmpdir.join('gzip.yaml'))):
        response = verify_pool_mgr.request('GET', url)
        assert_is_json(response.data)

    with vcr.use_cassette(str(tmpdir.join('gzip.yaml'))):
        assert_is_json(response.data)


def test_https_with_cert_validation_disabled(tmpdir, httpbin_secure, pool_mgr):
    with vcr.use_cassette(str(tmpdir.join('cert_validation_disabled.yaml'))):
        pool_mgr.request('GET', httpbin_secure.url)



import filecmp
import json
import shutil
import yaml

import vcr.migration


def test_try_migrate_with_json(tmpdir):
    cassette = tmpdir.join('cassette.json').strpath
    shutil.copy('tests/fixtures/migration/old_cassette.json', cassette)
    assert vcr.migration.try_migrate(cassette)
    with open('tests/fixtures/migration/new_cassette.json', 'r') as f:
        expected_json = json.load(f)
    with open(cassette, 'r') as f:
        actual_json = json.load(f)
    assert actual_json == expected_json


def test_try_migrate_with_yaml(tmpdir):
    cassette = tmpdir.join('cassette.yaml').strpath
    shutil.copy('tests/fixtures/migration/old_cassette.yaml', cassette)
    assert vcr.migration.try_migrate(cassette)
    with open('tests/fixtures/migration/new_cassette.yaml', 'r') as f:
        expected_yaml = yaml.load(f)
    with open(cassette, 'r') as f:
        actual_yaml = yaml.load(f)
    assert actual_yaml == expected_yaml


def test_try_migrate_with_invalid_or_new_cassettes(tmpdir):
    cassette = tmpdir.join('cassette').strpath
    files = [
        'tests/fixtures/migration/not_cassette.txt',
        'tests/fixtures/migration/new_cassette.yaml',
        'tests/fixtures/migration/new_cassette.json',
    ]
    for file_path in files:
        shutil.copy(file_path, cassette)
        assert not vcr.migration.try_migrate(cassette)
        assert filecmp.cmp(cassette, file_path)  # shold not change file



import vcr
from six.moves.urllib.request import urlopen


def true_matcher(r1, r2):
    return True


def false_matcher(r1, r2):
    return False


def test_registered_true_matcher(tmpdir, httpbin):
    my_vcr = vcr.VCR()
    my_vcr.register_matcher('true', true_matcher)
    testfile = str(tmpdir.join('test.yml'))
    with my_vcr.use_cassette(testfile, match_on=['true']):
        # These 2 different urls are stored as the same request
        urlopen(httpbin.url)
        urlopen(httpbin.url + '/get')

    with my_vcr.use_cassette(testfile, match_on=['true']):
        # I can get the response twice even though I only asked for it once
        urlopen(httpbin.url + '/get')
        urlopen(httpbin.url + '/get')


def test_registered_false_matcher(tmpdir, httpbin):
    my_vcr = vcr.VCR()
    my_vcr.register_matcher('false', false_matcher)
    testfile = str(tmpdir.join('test.yml'))
    with my_vcr.use_cassette(testfile, match_on=['false']) as cass:
        # These 2 different urls are stored as different requests
        urlopen(httpbin.url)
        urlopen(httpbin.url + '/get')
        assert len(cass) == 2



# -*- coding: utf-8 -*-
'''Integration tests with urllib2'''

from six.moves.urllib.request import urlopen
from six.moves.urllib_parse import urlencode
import pytest_httpbin.certs

# Internal imports
import vcr

from assertions import assert_cassette_has_one_response


def urlopen_with_cafile(*args, **kwargs):
    kwargs['cafile'] = pytest_httpbin.certs.where()
    try:
        return urlopen(*args, **kwargs)
    except TypeError:
        # python2/pypi don't let us override this
        del kwargs['cafile']
        return urlopen(*args, **kwargs)


def test_response_code(httpbin_both, tmpdir):
    '''Ensure we can read a response code from a fetch'''
    url = httpbin_both.url
    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        code = urlopen_with_cafile(url).getcode()

    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        assert code == urlopen_with_cafile(url).getcode()


def test_random_body(httpbin_both, tmpdir):
    '''Ensure we can read the content, and that it's served from cache'''
    url = httpbin_both.url + '/bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        body = urlopen_with_cafile(url).read()

    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        assert body == urlopen_with_cafile(url).read()


def test_response_headers(httpbin_both, tmpdir):
    '''Ensure we can get information from the response'''
    url = httpbin_both.url
    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        open1 = urlopen_with_cafile(url).info().items()

    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        open2 = urlopen_with_cafile(url).info().items()

        assert sorted(open1) == sorted(open2)


def test_effective_url(httpbin_both, tmpdir):
    '''Ensure that the effective_url is captured'''
    url = httpbin_both.url + '/redirect-to?url=/html'
    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        effective_url = urlopen_with_cafile(url).geturl()
        assert effective_url == httpbin_both.url + '/html'

    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        assert effective_url == urlopen_with_cafile(url).geturl()


def test_multiple_requests(httpbin_both, tmpdir):
    '''Ensure that we can cache multiple requests'''
    urls = [
        httpbin_both.url,
        httpbin_both.url,
        httpbin_both.url + '/get',
        httpbin_both.url + '/bytes/1024',
    ]
    with vcr.use_cassette(str(tmpdir.join('multiple.yaml'))) as cass:
        [urlopen_with_cafile(url) for url in urls]
    assert len(cass) == len(urls)


def test_get_data(httpbin_both, tmpdir):
    '''Ensure that it works with query data'''
    data = urlencode({'some': 1, 'data': 'here'})
    url = httpbin_both.url + '/get?' + data
    with vcr.use_cassette(str(tmpdir.join('get_data.yaml'))):
        res1 = urlopen_with_cafile(url).read()

    with vcr.use_cassette(str(tmpdir.join('get_data.yaml'))):
        res2 = urlopen_with_cafile(url).read()
    assert res1 == res2


def test_post_data(httpbin_both, tmpdir):
    '''Ensure that it works when posting data'''
    data = urlencode({'some': 1, 'data': 'here'}).encode('utf-8')
    url = httpbin_both.url + '/post'
    with vcr.use_cassette(str(tmpdir.join('post_data.yaml'))):
        res1 = urlopen_with_cafile(url, data).read()

    with vcr.use_cassette(str(tmpdir.join('post_data.yaml'))) as cass:
        res2 = urlopen_with_cafile(url, data).read()
        assert len(cass) == 1

    assert res1 == res2
    assert_cassette_has_one_response(cass)


def test_post_unicode_data(httpbin_both, tmpdir):
    '''Ensure that it works when posting unicode data'''
    data = urlencode({'snowman': u'☃'.encode('utf-8')}).encode('utf-8')
    url = httpbin_both.url + '/post'
    with vcr.use_cassette(str(tmpdir.join('post_data.yaml'))):
        res1 = urlopen_with_cafile(url, data).read()

    with vcr.use_cassette(str(tmpdir.join('post_data.yaml'))) as cass:
        res2 = urlopen_with_cafile(url, data).read()
        assert len(cass) == 1

    assert res1 == res2
    assert_cassette_has_one_response(cass)


def test_cross_scheme(tmpdir, httpbin_secure, httpbin):
    '''Ensure that requests between schemes are treated separately'''
    # First fetch a url under https, and then again under https and then
    # ensure that we haven't served anything out of cache, and we have two
    # requests / response pairs in the cassette
    with vcr.use_cassette(str(tmpdir.join('cross_scheme.yaml'))) as cass:
        urlopen_with_cafile(httpbin_secure.url)
        urlopen_with_cafile(httpbin.url)
        assert len(cass) == 2
        assert cass.play_count == 0


def test_decorator(httpbin_both, tmpdir):
    '''Test the decorator version of VCR.py'''
    url = httpbin_both.url

    @vcr.use_cassette(str(tmpdir.join('atts.yaml')))
    def inner1():
        return urlopen_with_cafile(url).getcode()

    @vcr.use_cassette(str(tmpdir.join('atts.yaml')))
    def inner2():
        return urlopen_with_cafile(url).getcode()

    assert inner1() == inner2()



import pytest
aiohttp = pytest.importorskip("aiohttp")

import asyncio  # NOQA
import sys  # NOQA

import aiohttp  # NOQA
import pytest  # NOQA
import vcr  # NOQA

from .aiohttp_utils import aiohttp_request  # NOQA


def get(url, as_text=True, **kwargs):
    loop = asyncio.get_event_loop()
    with aiohttp.ClientSession() as session:
        task = loop.create_task(aiohttp_request(session, 'GET', url, as_text, **kwargs))
        return loop.run_until_complete(task)


def post(url, as_text=True, **kwargs):
    loop = asyncio.get_event_loop()
    with aiohttp.ClientSession() as session:
        task = loop.create_task(aiohttp_request(session, 'POST', url, as_text, **kwargs))
        return loop.run_until_complete(task)


@pytest.fixture(params=["https", "http"])
def scheme(request):
    '''Fixture that returns both http and https.'''
    return request.param


def test_status(tmpdir, scheme):
    url = scheme + '://httpbin.org'
    with vcr.use_cassette(str(tmpdir.join('status.yaml'))):
        response, _ = get(url)

    with vcr.use_cassette(str(tmpdir.join('status.yaml'))) as cassette:
        cassette_response, _ = get(url)
        assert cassette_response.status == response.status
        assert cassette.play_count == 1


def test_headers(tmpdir, scheme):
    url = scheme + '://httpbin.org'
    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        response, _ = get(url)

    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))) as cassette:
        cassette_response, _ = get(url)
        assert cassette_response.headers == response.headers
        assert cassette.play_count == 1


def test_text(tmpdir, scheme):
    url = scheme + '://httpbin.org'
    with vcr.use_cassette(str(tmpdir.join('text.yaml'))):
        _, response_text = get(url)

    with vcr.use_cassette(str(tmpdir.join('text.yaml'))) as cassette:
        _, cassette_response_text = get(url)
        assert cassette_response_text == response_text
        assert cassette.play_count == 1


def test_json(tmpdir, scheme):
    url = scheme + '://httpbin.org/get'
    with vcr.use_cassette(str(tmpdir.join('json.yaml'))):
        _, response_json = get(url, as_text=False)

    with vcr.use_cassette(str(tmpdir.join('json.yaml'))) as cassette:
        _, cassette_response_json = get(url, as_text=False)
        assert cassette_response_json == response_json
        assert cassette.play_count == 1


def test_post(tmpdir, scheme):
    data = {'key1': 'value1', 'key2': 'value2'}
    url = scheme + '://httpbin.org/post'
    with vcr.use_cassette(str(tmpdir.join('post.yaml'))):
        _, response_json = post(url, data=data)

    with vcr.use_cassette(str(tmpdir.join('post.yaml'))) as cassette:
        _, cassette_response_json = post(url, data=data)
        assert cassette_response_json == response_json
        assert cassette.play_count == 1



import pytest
boto = pytest.importorskip("boto")

import boto  # NOQA
import boto.iam  # NOQA
from boto.s3.connection import S3Connection  # NOQA
from boto.s3.key import Key  # NOQA
import vcr  # NOQA
try:  # NOQA
    from ConfigParser import DuplicateSectionError  # NOQA
except ImportError:  # NOQA
    # python3
    from configparser import DuplicateSectionError  # NOQA


def test_boto_stubs(tmpdir):
    with vcr.use_cassette(str(tmpdir.join('boto-stubs.yml'))):
        # Perform the imports within the patched context so that
        # CertValidatingHTTPSConnection refers to the patched version.
        from boto.https_connection import CertValidatingHTTPSConnection
        from vcr.stubs.boto_stubs import VCRCertValidatingHTTPSConnection
        # Prove that the class was patched by the stub and that we can instantiate it.
        assert issubclass(CertValidatingHTTPSConnection, VCRCertValidatingHTTPSConnection)
        CertValidatingHTTPSConnection('hostname.does.not.matter')


def test_boto_without_vcr():
    s3_conn = S3Connection()
    s3_bucket = s3_conn.get_bucket('boto-demo-1394171994')  # a bucket you can access
    k = Key(s3_bucket)
    k.key = 'test.txt'
    k.set_contents_from_string('hello world i am a string')


def test_boto_medium_difficulty(tmpdir):
    s3_conn = S3Connection()
    s3_bucket = s3_conn.get_bucket('boto-demo-1394171994')  # a bucket you can access
    with vcr.use_cassette(str(tmpdir.join('boto-medium.yml'))):
        k = Key(s3_bucket)
        k.key = 'test.txt'
        k.set_contents_from_string('hello world i am a string')

    with vcr.use_cassette(str(tmpdir.join('boto-medium.yml'))):
        k = Key(s3_bucket)
        k.key = 'test.txt'
        k.set_contents_from_string('hello world i am a string')


def test_boto_hardcore_mode(tmpdir):
    with vcr.use_cassette(str(tmpdir.join('boto-hardcore.yml'))):
        s3_conn = S3Connection()
        s3_bucket = s3_conn.get_bucket('boto-demo-1394171994')  # a bucket you can access
        k = Key(s3_bucket)
        k.key = 'test.txt'
        k.set_contents_from_string('hello world i am a string')

    with vcr.use_cassette(str(tmpdir.join('boto-hardcore.yml'))):
        s3_conn = S3Connection()
        s3_bucket = s3_conn.get_bucket('boto-demo-1394171994')  # a bucket you can access
        k = Key(s3_bucket)
        k.key = 'test.txt'
        k.set_contents_from_string('hello world i am a string')


def test_boto_iam(tmpdir):
    try:
        boto.config.add_section('Boto')
    except DuplicateSectionError:
        pass
    # Ensure that boto uses HTTPS
    boto.config.set('Boto', 'is_secure', 'true')
    # Ensure that boto uses CertValidatingHTTPSConnection
    boto.config.set('Boto', 'https_validate_certificates', 'true')

    with vcr.use_cassette(str(tmpdir.join('boto-iam.yml'))):
        iam_conn = boto.iam.connect_to_region('universal')
        iam_conn.get_all_users()

    with vcr.use_cassette(str(tmpdir.join('boto-iam.yml'))):
        iam_conn = boto.iam.connect_to_region('universal')
        iam_conn.get_all_users()



# coding: UTF-8
from vcr.stubs import VCRHTTPResponse


def test_response_should_have_headers_field():
    recorded_response = {
        "status": {
            "message": "OK",
            "code": 200
        },
        "headers": {
            "content-length": ["0"],
            "server": ["gunicorn/18.0"],
            "connection": ["Close"],
            "access-control-allow-credentials": ["true"],
            "date": ["Fri, 24 Oct 2014 18:35:37 GMT"],
            "access-control-allow-origin": ["*"],
            "content-type": ["text/html; charset=utf-8"],
        },
        "body": {
            "string": b""
        }
    }
    response = VCRHTTPResponse(recorded_response)

    assert response.headers is not None


def test_response_headers_should_be_equal_to_msg():
    recorded_response = {
        "status": {
            "message": b"OK",
            "code": 200
        },
        "headers": {
            "content-length": ["0"],
            "server": ["gunicorn/18.0"],
            "connection": ["Close"],
            "content-type": ["text/html; charset=utf-8"],
        },
        "body": {
            "string": b""
        }
    }
    response = VCRHTTPResponse(recorded_response)

    assert response.headers == response.msg


def test_response_headers_should_have_correct_values():
    recorded_response = {
        "status": {
            "message": "OK",
            "code": 200
        },
        "headers": {
            "content-length": ["10806"],
            "date": ["Fri, 24 Oct 2014 18:35:37 GMT"],
            "content-type": ["text/html; charset=utf-8"],
        },
        "body": {
            "string": b""
        }
    }
    response = VCRHTTPResponse(recorded_response)

    assert response.headers.get('content-length') == "10806"
    assert response.headers.get('date') == "Fri, 24 Oct 2014 18:35:37 GMT"



import pytest
import vcr
from six.moves.urllib.request import urlopen


def test_making_extra_request_raises_exception(tmpdir, httpbin):
    # make two requests in the first request that are considered
    # identical (since the match is based on method)
    with vcr.use_cassette(str(tmpdir.join('test.json')), match_on=['method']):
        urlopen(httpbin.url + '/status/200')
        urlopen(httpbin.url + '/status/201')

    # Now, try to make three requests.  The first two should return the
    # correct status codes in order, and the third should raise an
    # exception.
    with vcr.use_cassette(str(tmpdir.join('test.json')), match_on=['method']):
        assert urlopen(httpbin.url + '/status/200').getcode() == 200
        assert urlopen(httpbin.url + '/status/201').getcode() == 201
        with pytest.raises(Exception):
            urlopen(httpbin.url + '/status/200')



import pytest

from vcr.request import Request, HeadersDict


def test_str():
    req = Request('GET', 'http://www.google.com/', '', {})
    str(req) == '<Request (GET) http://www.google.com/>'


def test_headers():
    headers = {'X-Header1': ['h1'], 'X-Header2': 'h2'}
    req = Request('GET', 'http://go.com/', '', headers)
    assert req.headers == {'X-Header1': 'h1', 'X-Header2': 'h2'}
    req.headers['X-Header1'] = 'h11'
    assert req.headers == {'X-Header1': 'h11', 'X-Header2': 'h2'}


def test_add_header_deprecated():
    req = Request('GET', 'http://go.com/', '', {})
    pytest.deprecated_call(req.add_header, 'foo', 'bar')
    assert req.headers == {'foo': 'bar'}


@pytest.mark.parametrize("uri, expected_port", [
    ('http://go.com/', 80),
    ('http://go.com:80/', 80),
    ('http://go.com:3000/', 3000),
    ('https://go.com/', 443),
    ('https://go.com:443/', 443),
    ('https://go.com:3000/', 3000),
])
def test_port(uri, expected_port):
    req = Request('GET', uri, '', {})
    assert req.port == expected_port


def test_uri():
    req = Request('GET', 'http://go.com/', '', {})
    assert req.uri == 'http://go.com/'

    req = Request('GET', 'http://go.com:80/', '', {})
    assert req.uri == 'http://go.com:80/'


def test_HeadersDict():

    # Simple test of CaseInsensitiveDict
    h = HeadersDict()
    assert h == {}
    h['Content-Type'] = 'application/json'
    assert h == {'Content-Type': 'application/json'}
    assert h['content-type'] == 'application/json'
    assert h['CONTENT-TYPE'] == 'application/json'

    # Test feature of HeadersDict: devolve list to first element
    h = HeadersDict()
    assert h == {}
    h['x'] = ['foo', 'bar']
    assert h == {'x': 'foo'}

    # Test feature of HeadersDict: preserve original key case
    h = HeadersDict()
    assert h == {}
    h['Content-Type'] = 'application/json'
    assert h == {'Content-Type': 'application/json'}
    h['content-type'] = 'text/plain'
    assert h == {'Content-Type': 'text/plain'}
    h['CONtent-tyPE'] = 'whoa'
    assert h == {'Content-Type': 'whoa'}



import vcr
from six.moves.urllib.request import urlopen


def test_recorded_request_uri_with_redirected_request(tmpdir, httpbin):
    with vcr.use_cassette(str(tmpdir.join('test.yml'))) as cass:
        assert len(cass) == 0
        urlopen(httpbin.url + '/redirect/3')
        assert cass.requests[0].uri == httpbin.url + '/redirect/3'
        assert cass.requests[3].uri == httpbin.url + '/get'
        assert len(cass) == 4


def test_records_multiple_header_values(tmpdir, httpbin):
    with vcr.use_cassette(str(tmpdir.join('test.yml'))) as cass:
        assert len(cass) == 0
        urlopen(httpbin.url + '/response-headers?foo=bar&foo=baz')
        assert len(cass) == 1
        assert cass.responses[0]['headers']['foo'] == ['bar', 'baz']



import pytest

import vcr.persist
from vcr.serializers import jsonserializer, yamlserializer


@pytest.mark.parametrize("cassette_path, serializer", [
    ('tests/fixtures/migration/old_cassette.json', jsonserializer),
    ('tests/fixtures/migration/old_cassette.yaml', yamlserializer),
])
def test_load_cassette_with_old_cassettes(cassette_path, serializer):
    with pytest.raises(ValueError) as excinfo:
        vcr.persist.load_cassette(cassette_path, serializer)
    assert "run the migration script" in excinfo.exconly()


@pytest.mark.parametrize("cassette_path, serializer", [
    ('tests/fixtures/migration/not_cassette.txt', jsonserializer),
    ('tests/fixtures/migration/not_cassette.txt', yamlserializer),
])
def test_load_cassette_with_invalid_cassettes(cassette_path, serializer):
    with pytest.raises(Exception) as excinfo:
        vcr.persist.load_cassette(cassette_path, serializer)
    assert "run the migration script" not in excinfo.exconly()



# -*- encoding: utf-8 -*-
import pytest

from vcr.compat import mock
from vcr.request import Request
from vcr.serialize import deserialize, serialize
from vcr.serializers import yamlserializer, jsonserializer


def test_deserialize_old_yaml_cassette():
    with open('tests/fixtures/migration/old_cassette.yaml', 'r') as f:
        with pytest.raises(ValueError):
            deserialize(f.read(), yamlserializer)


def test_deserialize_old_json_cassette():
    with open('tests/fixtures/migration/old_cassette.json', 'r') as f:
        with pytest.raises(ValueError):
            deserialize(f.read(), jsonserializer)


def test_deserialize_new_yaml_cassette():
    with open('tests/fixtures/migration/new_cassette.yaml', 'r') as f:
        deserialize(f.read(), yamlserializer)


def test_deserialize_new_json_cassette():
    with open('tests/fixtures/migration/new_cassette.json', 'r') as f:
        deserialize(f.read(), jsonserializer)


REQBODY_TEMPLATE = u'''\
interactions:
- request:
    body: {req_body}
    headers:
      Content-Type: [application/x-www-form-urlencoded]
      Host: [httpbin.org]
    method: POST
    uri: http://httpbin.org/post
  response:
    body: {{string: ""}}
    headers:
      content-length: ['0']
      content-type: [application/json]
    status: {{code: 200, message: OK}}
'''


# A cassette generated under Python 2 stores the request body as a string,
# but the same cassette generated under Python 3 stores it as "!!binary".
# Make sure we accept both forms, regardless of whether we're running under
# Python 2 or 3.
@pytest.mark.parametrize("req_body, expect", [
    # Cassette written under Python 2 (pure ASCII body)
    ('x=5&y=2', b'x=5&y=2'),
    # Cassette written under Python 3 (pure ASCII body)
    ('!!binary |\n      eD01Jnk9Mg==', b'x=5&y=2'),

    # Request body has non-ASCII chars (x=föo&y=2), encoded in UTF-8.
    ('!!python/str "x=f\\xF6o&y=2"', b'x=f\xc3\xb6o&y=2'),
    ('!!binary |\n      eD1mw7ZvJnk9Mg==', b'x=f\xc3\xb6o&y=2'),

    # Same request body, this time encoded in UTF-16. In this case, we
    # write the same YAML file under both Python 2 and 3, so there's only
    # one test case here.
    ('!!binary |\n      //54AD0AZgD2AG8AJgB5AD0AMgA=',
     b'\xff\xfex\x00=\x00f\x00\xf6\x00o\x00&\x00y\x00=\x002\x00'),

    # Same again, this time encoded in ISO-8859-1.
    ('!!binary |\n      eD1m9m8meT0y', b'x=f\xf6o&y=2'),
])
def test_deserialize_py2py3_yaml_cassette(tmpdir, req_body, expect):
    cfile = tmpdir.join('test_cassette.yaml')
    cfile.write(REQBODY_TEMPLATE.format(req_body=req_body))
    with open(str(cfile)) as f:
        (requests, responses) = deserialize(f.read(), yamlserializer)
    assert requests[0].body == expect


@mock.patch.object(jsonserializer.json, 'dumps',
                   side_effect=UnicodeDecodeError('utf-8', b'unicode error in serialization',
                                                  0, 10, 'blew up'))
def test_serialize_constructs_UnicodeDecodeError(mock_dumps):
    with pytest.raises(UnicodeDecodeError):
        jsonserializer.serialize({})


def test_serialize_empty_request():
    request = Request(
        method='POST',
        uri='http://localhost/',
        body='',
        headers={},
    )

    serialize(
        {'requests': [request], 'responses': [{}]},
        jsonserializer
    )


def test_serialize_json_request():
    request = Request(
        method='POST',
        uri='http://localhost/',
        body="{'hello': 'world'}",
        headers={},
    )

    serialize(
        {'requests': [request], 'responses': [{}]},
        jsonserializer
    )


def test_serialize_binary_request():
    msg = "Does this HTTP interaction contain binary data?"

    request = Request(
        method='POST',
        uri='http://localhost/',
        body=b'\x8c',
        headers={},
    )

    try:
        serialize(
            {'requests': [request], 'responses': [{}]},
            jsonserializer
        )
    except (UnicodeDecodeError, TypeError) as exc:
        assert msg in str(exc)



import copy
import inspect
import os

from six.moves import http_client as httplib
import pytest
import yaml

from vcr.compat import mock, contextlib
from vcr.cassette import Cassette
from vcr.errors import UnhandledHTTPRequestError
from vcr.patch import force_reset
from vcr.stubs import VCRHTTPSConnection


def test_cassette_load(tmpdir):
    a_file = tmpdir.join('test_cassette.yml')
    a_file.write(yaml.dump({'interactions': [
        {'request': {'body': '', 'uri': 'foo', 'method': 'GET', 'headers': {}},
         'response': 'bar'}
    ]}))
    a_cassette = Cassette.load(path=str(a_file))
    assert len(a_cassette) == 1


def test_cassette_not_played():
    a = Cassette('test')
    assert not a.play_count


def test_cassette_append():
    a = Cassette('test')
    a.append('foo', 'bar')
    assert a.requests == ['foo']
    assert a.responses == ['bar']


def test_cassette_len():
    a = Cassette('test')
    a.append('foo', 'bar')
    a.append('foo2', 'bar2')
    assert len(a) == 2


def _mock_requests_match(request1, request2, matchers):
    return request1 == request2


@mock.patch('vcr.cassette.requests_match', _mock_requests_match)
def test_cassette_contains():
    a = Cassette('test')
    a.append('foo', 'bar')
    assert 'foo' in a


@mock.patch('vcr.cassette.requests_match', _mock_requests_match)
def test_cassette_responses_of():
    a = Cassette('test')
    a.append('foo', 'bar')
    assert a.responses_of('foo') == ['bar']


@mock.patch('vcr.cassette.requests_match', _mock_requests_match)
def test_cassette_get_missing_response():
    a = Cassette('test')
    with pytest.raises(UnhandledHTTPRequestError):
        a.responses_of('foo')


@mock.patch('vcr.cassette.requests_match', _mock_requests_match)
def test_cassette_cant_read_same_request_twice():
    a = Cassette('test')
    a.append('foo', 'bar')
    a.play_response('foo')
    with pytest.raises(UnhandledHTTPRequestError):
        a.play_response('foo')


def make_get_request():
    conn = httplib.HTTPConnection("www.python.org")
    conn.request("GET", "/index.html")
    return conn.getresponse()


@mock.patch('vcr.cassette.requests_match', return_value=True)
@mock.patch('vcr.cassette.load_cassette', lambda *args, **kwargs: (('foo',), (mock.MagicMock(),)))
@mock.patch('vcr.cassette.Cassette.can_play_response_for', return_value=True)
@mock.patch('vcr.stubs.VCRHTTPResponse')
def test_function_decorated_with_use_cassette_can_be_invoked_multiple_times(*args):
    decorated_function = Cassette.use(path='test')(make_get_request)
    for i in range(4):
        decorated_function()


def test_arg_getter_functionality():
    arg_getter = mock.Mock(return_value={'path': 'test'})
    context_decorator = Cassette.use_arg_getter(arg_getter)

    with context_decorator as cassette:
        assert cassette._path == 'test'

    arg_getter.return_value = {'path': 'other'}

    with context_decorator as cassette:
        assert cassette._path == 'other'

    arg_getter.return_value = {'path': 'other', 'filter_headers': ('header_name',)}

    @context_decorator
    def function():
        pass

    with mock.patch.object(
        Cassette, 'load',
        return_value=mock.MagicMock(inject=False)
    ) as cassette_load:
        function()
        cassette_load.assert_called_once_with(**arg_getter.return_value)


def test_cassette_not_all_played():
    a = Cassette('test')
    a.append('foo', 'bar')
    assert not a.all_played


@mock.patch('vcr.cassette.requests_match', _mock_requests_match)
def test_cassette_all_played():
    a = Cassette('test')
    a.append('foo', 'bar')
    a.play_response('foo')
    assert a.all_played


def test_before_record_response():
    before_record_response = mock.Mock(return_value='mutated')
    cassette = Cassette('test', before_record_response=before_record_response)
    cassette.append('req', 'res')

    before_record_response.assert_called_once_with('res')
    assert cassette.responses[0] == 'mutated'


def assert_get_response_body_is(value):
    conn = httplib.HTTPConnection("www.python.org")
    conn.request("GET", "/index.html")
    assert conn.getresponse().read().decode('utf8') == value


@mock.patch('vcr.cassette.requests_match', _mock_requests_match)
@mock.patch('vcr.cassette.Cassette.can_play_response_for', return_value=True)
@mock.patch('vcr.cassette.Cassette._save', return_value=True)
def test_nesting_cassette_context_managers(*args):
    first_response = {'body': {'string': b'first_response'}, 'headers': {},
                      'status': {'message': 'm', 'code': 200}}

    second_response = copy.deepcopy(first_response)
    second_response['body']['string'] = b'second_response'

    with contextlib.ExitStack() as exit_stack:
        first_cassette = exit_stack.enter_context(Cassette.use(path='test'))
        exit_stack.enter_context(mock.patch.object(first_cassette, 'play_response',
                                                   return_value=first_response))
        assert_get_response_body_is('first_response')

        # Make sure a second cassette can supercede the first
        with Cassette.use(path='test') as second_cassette:
            with mock.patch.object(second_cassette, 'play_response', return_value=second_response):
                assert_get_response_body_is('second_response')

        # Now the first cassette should be back in effect
        assert_get_response_body_is('first_response')


def test_nesting_context_managers_by_checking_references_of_http_connection():
    original = httplib.HTTPConnection
    with Cassette.use(path='test'):
        first_cassette_HTTPConnection = httplib.HTTPConnection
        with Cassette.use(path='test'):
            second_cassette_HTTPConnection = httplib.HTTPConnection
            assert second_cassette_HTTPConnection is not first_cassette_HTTPConnection
            with Cassette.use(path='test'):
                assert httplib.HTTPConnection is not second_cassette_HTTPConnection
                with force_reset():
                    assert httplib.HTTPConnection is original
            assert httplib.HTTPConnection is second_cassette_HTTPConnection
        assert httplib.HTTPConnection is first_cassette_HTTPConnection


def test_custom_patchers():
    class Test(object):
        attribute = None
    with Cassette.use(path='custom_patches',
                      custom_patches=((Test, 'attribute', VCRHTTPSConnection),)):
        assert issubclass(Test.attribute, VCRHTTPSConnection)
        assert VCRHTTPSConnection is not Test.attribute
        old_attribute = Test.attribute

        with Cassette.use(path='custom_patches',
                          custom_patches=((Test, 'attribute', VCRHTTPSConnection),)):
            assert issubclass(Test.attribute, VCRHTTPSConnection)
            assert VCRHTTPSConnection is not Test.attribute
            assert Test.attribute is not old_attribute

        assert issubclass(Test.attribute, VCRHTTPSConnection)
        assert VCRHTTPSConnection is not Test.attribute
        assert Test.attribute is old_attribute


def test_decorated_functions_are_reentrant():
    info = {"second": False}
    original_conn = httplib.HTTPConnection

    @Cassette.use(path='whatever', inject=True)
    def test_function(cassette):
        if info['second']:
            assert httplib.HTTPConnection is not info['first_conn']
        else:
            info['first_conn'] = httplib.HTTPConnection
            info['second'] = True
            test_function()
            assert httplib.HTTPConnection is info['first_conn']

    test_function()
    assert httplib.HTTPConnection is original_conn


def test_cassette_use_called_without_path_uses_function_to_generate_path():
    @Cassette.use(inject=True)
    def function_name(cassette):
        assert cassette._path == 'function_name'
    function_name()


def test_path_transformer_with_function_path():
    def path_transformer(path):
        return os.path.join('a', path)

    @Cassette.use(inject=True, path_transformer=path_transformer)
    def function_name(cassette):
        assert cassette._path == os.path.join('a', 'function_name')

    function_name()


def test_path_transformer_with_context_manager():
    with Cassette.use(
        path='b', path_transformer=lambda *args: 'a'
    ) as cassette:
        assert cassette._path == 'a'


def test_path_transformer_None():
    with Cassette.use(
        path='a', path_transformer=None,
    ) as cassette:
        assert cassette._path == 'a'


def test_func_path_generator():
    def generator(function):
        return os.path.join(os.path.dirname(inspect.getfile(function)),
                            function.__name__)

    @Cassette.use(inject=True, func_path_generator=generator)
    def function_name(cassette):
        assert cassette._path == os.path.join(os.path.dirname(__file__), 'function_name')

    function_name()


def test_use_as_decorator_on_coroutine():
    original_http_connetion = httplib.HTTPConnection

    @Cassette.use(inject=True)
    def test_function(cassette):
        assert httplib.HTTPConnection.cassette is cassette
        assert httplib.HTTPConnection is not original_http_connetion
        value = yield 1
        assert value == 1
        assert httplib.HTTPConnection.cassette is cassette
        assert httplib.HTTPConnection is not original_http_connetion
        value = yield 2
        assert value == 2

    coroutine = test_function()
    value = next(coroutine)
    while True:
        try:
            value = coroutine.send(value)
        except StopIteration:
            break


def test_use_as_decorator_on_generator():
    original_http_connetion = httplib.HTTPConnection

    @Cassette.use(inject=True)
    def test_function(cassette):
        assert httplib.HTTPConnection.cassette is cassette
        assert httplib.HTTPConnection is not original_http_connetion
        yield 1
        assert httplib.HTTPConnection.cassette is cassette
        assert httplib.HTTPConnection is not original_http_connetion
        yield 2

    assert list(test_function()) == [1, 2]



import asyncio


@asyncio.coroutine
def aiohttp_request(session, method, url, as_text, **kwargs):
    response = yield from session.request(method, url, **kwargs)  # NOQA: E999
    return response, (yield from response.text()) if as_text else (yield from response.json())  # NOQA: E999



import vcr
import zlib
import six.moves.http_client as httplib

from assertions import assert_is_json


def _headers_are_case_insensitive(host, port):
    conn = httplib.HTTPConnection(host, port)
    conn.request('GET', "/cookies/set?k1=v1")
    r1 = conn.getresponse()
    cookie_data1 = r1.getheader('set-cookie')
    conn = httplib.HTTPConnection(host, port)
    conn.request('GET', "/cookies/set?k1=v1")
    r2 = conn.getresponse()
    cookie_data2 = r2.getheader('Set-Cookie')
    return cookie_data1 == cookie_data2


def test_case_insensitivity(tmpdir, httpbin):
    testfile = str(tmpdir.join('case_insensitivity.yml'))
    # check if headers are case insensitive outside of vcrpy
    host, port = httpbin.host, httpbin.port
    outside = _headers_are_case_insensitive(host, port)
    with vcr.use_cassette(testfile):
        # check if headers are case insensitive inside of vcrpy
        inside = _headers_are_case_insensitive(host, port)
        # check if headers are case insensitive after vcrpy deserializes headers
        inside2 = _headers_are_case_insensitive(host, port)

    # behavior should be the same both inside and outside
    assert outside == inside == inside2


def _multiple_header_value(httpbin):
    conn = httplib.HTTPConnection(httpbin.host, httpbin.port)
    conn.request('GET', "/response-headers?foo=bar&foo=baz")
    r = conn.getresponse()
    return r.getheader('foo')


def test_multiple_headers(tmpdir, httpbin):
    testfile = str(tmpdir.join('multiple_headers.yaml'))
    outside = _multiple_header_value(httpbin)

    with vcr.use_cassette(testfile):
        inside = _multiple_header_value(httpbin)

    assert outside == inside


def test_original_decoded_response_is_not_modified(tmpdir, httpbin):
    testfile = str(tmpdir.join('decoded_response.yml'))
    host, port = httpbin.host, httpbin.port

    conn = httplib.HTTPConnection(host, port)
    conn.request('GET', '/gzip')
    outside = conn.getresponse()

    with vcr.use_cassette(testfile, decode_compressed_response=True):
        conn = httplib.HTTPConnection(host, port)
        conn.request('GET', '/gzip')
        inside = conn.getresponse()

        # Assert that we do not modify the original response while appending
        # to the casssette.
        assert 'gzip' == inside.headers['content-encoding']

        # They should effectively be the same response.
        inside_headers = (h for h in inside.headers.items() if h[0].lower() != 'date')
        outside_headers = (h for h in outside.getheaders() if h[0].lower() != 'date')
        assert set(inside_headers) == set(outside_headers)
        inside = zlib.decompress(inside.read(), 16+zlib.MAX_WBITS)
        outside = zlib.decompress(outside.read(), 16+zlib.MAX_WBITS)
        assert inside == outside

    # Even though the above are raw bytes, the JSON data should have been
    # decoded and saved to the cassette.
    with vcr.use_cassette(testfile):
        conn = httplib.HTTPConnection(host, port)
        conn.request('GET', '/gzip')
        inside = conn.getresponse()

        assert 'content-encoding' not in inside.headers
        assert_is_json(inside.read())



# -*- coding: utf-8 -*-
'''Test requests' interaction with vcr'''
import pytest
import vcr
from assertions import assert_cassette_empty, assert_is_json


requests = pytest.importorskip("requests")


def test_status_code(httpbin_both, tmpdir):
    '''Ensure that we can read the status code'''
    url = httpbin_both.url + '/'
    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        status_code = requests.get(url).status_code

    with vcr.use_cassette(str(tmpdir.join('atts.yaml'))):
        assert status_code == requests.get(url).status_code


def test_headers(httpbin_both, tmpdir):
    '''Ensure that we can read the headers back'''
    url = httpbin_both + '/'
    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        headers = requests.get(url).headers

    with vcr.use_cassette(str(tmpdir.join('headers.yaml'))):
        assert headers == requests.get(url).headers


def test_body(tmpdir, httpbin_both):
    '''Ensure the responses are all identical enough'''
    url = httpbin_both + '/bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        content = requests.get(url).content

    with vcr.use_cassette(str(tmpdir.join('body.yaml'))):
        assert content == requests.get(url).content


def test_effective_url(tmpdir, httpbin_both):
    '''Ensure that the effective_url is captured'''
    url = httpbin_both.url + '/redirect-to?url=/html'
    with vcr.use_cassette(str(tmpdir.join('url.yaml'))):
        effective_url = requests.get(url).url
        assert effective_url == httpbin_both.url + '/html'

    with vcr.use_cassette(str(tmpdir.join('url.yaml'))):
        assert effective_url == requests.get(url).url


def test_auth(tmpdir, httpbin_both):
    '''Ensure that we can handle basic auth'''
    auth = ('user', 'passwd')
    url = httpbin_both + '/basic-auth/user/passwd'
    with vcr.use_cassette(str(tmpdir.join('auth.yaml'))):
        one = requests.get(url, auth=auth)

    with vcr.use_cassette(str(tmpdir.join('auth.yaml'))):
        two = requests.get(url, auth=auth)
        assert one.content == two.content
        assert one.status_code == two.status_code


def test_auth_failed(tmpdir, httpbin_both):
    '''Ensure that we can save failed auth statuses'''
    auth = ('user', 'wrongwrongwrong')
    url = httpbin_both + '/basic-auth/user/passwd'
    with vcr.use_cassette(str(tmpdir.join('auth-failed.yaml'))) as cass:
        # Ensure that this is empty to begin with
        assert_cassette_empty(cass)
        one = requests.get(url, auth=auth)
        two = requests.get(url, auth=auth)
        assert one.content == two.content
        assert one.status_code == two.status_code == 401


def test_post(tmpdir, httpbin_both):
    '''Ensure that we can post and cache the results'''
    data = {'key1': 'value1', 'key2': 'value2'}
    url = httpbin_both + '/post'
    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        req1 = requests.post(url, data).content

    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        req2 = requests.post(url, data).content

    assert req1 == req2


def test_post_chunked_binary(tmpdir, httpbin_both):
    '''Ensure that we can send chunked binary without breaking while trying to concatenate bytes with str.'''
    data1 = iter([b'data', b'to', b'send'])
    data2 = iter([b'data', b'to', b'send'])
    url = httpbin_both.url + '/post'
    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        req1 = requests.post(url, data1).content
        print(req1)

    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        req2 = requests.post(url, data2).content

    assert req1 == req2


def test_redirects(tmpdir, httpbin_both):
    '''Ensure that we can handle redirects'''
    url = httpbin_both + '/redirect-to?url=bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))):
        content = requests.get(url).content

    with vcr.use_cassette(str(tmpdir.join('requests.yaml'))) as cass:
        assert content == requests.get(url).content
        # Ensure that we've now cached *two* responses. One for the redirect
        # and one for the final fetch
        assert len(cass) == 2
        assert cass.play_count == 2


def test_cross_scheme(tmpdir, httpbin_secure, httpbin):
    '''Ensure that requests between schemes are treated separately'''
    # First fetch a url under http, and then again under https and then
    # ensure that we haven't served anything out of cache, and we have two
    # requests / response pairs in the cassette
    with vcr.use_cassette(str(tmpdir.join('cross_scheme.yaml'))) as cass:
        requests.get(httpbin_secure + '/')
        requests.get(httpbin + '/')
        assert cass.play_count == 0
        assert len(cass) == 2


def test_gzip(tmpdir, httpbin_both):
    '''
    Ensure that requests (actually urllib3) is able to automatically decompress
    the response body
    '''
    url = httpbin_both + '/gzip'
    response = requests.get(url)

    with vcr.use_cassette(str(tmpdir.join('gzip.yaml'))):
        response = requests.get(url)
        assert_is_json(response.content)

    with vcr.use_cassette(str(tmpdir.join('gzip.yaml'))):
        assert_is_json(response.content)


def test_session_and_connection_close(tmpdir, httpbin):
    '''
    This tests the issue in https://github.com/kevin1024/vcrpy/issues/48

    If you use a requests.session and the connection is closed, then an
    exception is raised in the urllib3 module vendored into requests:
    `AttributeError: 'NoneType' object has no attribute 'settimeout'`
    '''
    with vcr.use_cassette(str(tmpdir.join('session_connection_closed.yaml'))):
        session = requests.session()

        session.get(httpbin + '/get', headers={'Connection': 'close'})
        session.get(httpbin + '/get', headers={'Connection': 'close'})


def test_https_with_cert_validation_disabled(tmpdir, httpbin_secure):
    with vcr.use_cassette(str(tmpdir.join('cert_validation_disabled.yaml'))):
        requests.get(httpbin_secure.url, verify=False)


def test_session_can_make_requests_after_requests_unpatched(tmpdir, httpbin):
    with vcr.use_cassette(str(tmpdir.join('test_session_after_unpatched.yaml'))):
        session = requests.session()
        session.get(httpbin + '/get')

    with vcr.use_cassette(str(tmpdir.join('test_session_after_unpatched.yaml'))):
        session = requests.session()
        session.get(httpbin + '/get')

    session.get(httpbin + '/status/200')


def test_session_created_before_use_cassette_is_patched(tmpdir, httpbin_both):
    url = httpbin_both + '/bytes/1024'
    # Record arbitrary, random data to the cassette
    with vcr.use_cassette(str(tmpdir.join('session_created_outside.yaml'))):
        session = requests.session()
        body = session.get(url).content

    # Create a session outside of any cassette context manager
    session = requests.session()
    # Make a request to make sure that a connectionpool is instantiated
    session.get(httpbin_both + '/get')

    with vcr.use_cassette(str(tmpdir.join('session_created_outside.yaml'))):
        # These should only be the same if the patching succeeded.
        assert session.get(url).content == body


def test_nested_cassettes_with_session_created_before_nesting(httpbin_both, tmpdir):
    '''
    This tests ensures that a session that was created while one cassette was
    active is patched to the use the responses of a second cassette when it
    is enabled.
    '''
    url = httpbin_both + '/bytes/1024'
    with vcr.use_cassette(str(tmpdir.join('first_nested.yaml'))):
        session = requests.session()
        first_body = session.get(url).content
        with vcr.use_cassette(str(tmpdir.join('second_nested.yaml'))):
            second_body = session.get(url).content
            third_body = requests.get(url).content

    with vcr.use_cassette(str(tmpdir.join('second_nested.yaml'))):
        session = requests.session()
        assert session.get(url).content == second_body
        with vcr.use_cassette(str(tmpdir.join('first_nested.yaml'))):
            assert session.get(url).content == first_body
        assert session.get(url).content == third_body

    # Make sure that the session can now get content normally.
    assert 'User-agent' in session.get(httpbin_both.url + '/robots.txt').text


def test_post_file(tmpdir, httpbin_both):
    '''Ensure that we handle posting a file.'''
    url = httpbin_both + '/post'
    with vcr.use_cassette(str(tmpdir.join('post_file.yaml'))) as cass:
        # Don't use 2.7+ only style ',' separated with here because we support python 2.6
        with open('tox.ini') as f:
            original_response = requests.post(url, f).content

    # This also tests that we do the right thing with matching the body when they are files.
    with vcr.use_cassette(str(tmpdir.join('post_file.yaml')),
                          match_on=('method', 'scheme', 'host', 'port', 'path', 'query', 'body')) as cass:
        with open('tox.ini', 'rb') as f:
            tox_content = f.read()
        assert cass.requests[0].body.read() == tox_content
        with open('tox.ini', 'rb') as f:
            new_response = requests.post(url, f).content
        assert original_response == new_response


def test_filter_post_params(tmpdir, httpbin_both):
    '''
    This tests the issue in https://github.com/kevin1024/vcrpy/issues/158

    Ensure that a post request made through requests can still be filtered.
    with vcr.use_cassette(cass_file, filter_post_data_parameters=['id']) as cass:
        assert b'id=secret' not in cass.requests[0].body
    '''
    url = httpbin_both.url + '/post'
    cass_loc = str(tmpdir.join('filter_post_params.yaml'))
    with vcr.use_cassette(cass_loc, filter_post_data_parameters=['key']) as cass:
        requests.post(url, data={'key': 'value'})
    with vcr.use_cassette(cass_loc, filter_post_data_parameters=['key']) as cass:
        assert b'key=value' not in cass.requests[0].body



from .persisters.filesystem import FilesystemPersister
from .serialize import serialize, deserialize


def load_cassette(cassette_path, serializer):
    with open(cassette_path) as f:
        cassette_content = f.read()
        cassette = deserialize(cassette_content, serializer)
        return cassette


def save_cassette(cassette_path, cassette_dict, serializer):
    data = serialize(cassette_dict, serializer)
    FilesystemPersister.write(cassette_path, data)



import pytest
boto3 = pytest.importorskip("boto3")

import boto3  # NOQA
import vcr  # NOQA

bucket = 'boto3-demo-1337'              # a bucket you can access
key = 'test/my_test.txt'                # key with r+w access
content = 'hello world i am a string'   # content to put in the test file


def test_boto_stubs(tmpdir):
    with vcr.use_cassette(str(tmpdir.join('boto3-stubs.yml'))):
        # Perform the imports within the patched context so that
        # HTTPConnection, VerifiedHTTPSConnection refers to the patched version.
        from botocore.vendored.requests.packages.urllib3.connectionpool import \
            HTTPConnection, VerifiedHTTPSConnection
        from vcr.stubs.boto3_stubs import VCRRequestsHTTPConnection, VCRRequestsHTTPSConnection
        # Prove that the class was patched by the stub and that we can instantiate it.
        assert issubclass(HTTPConnection, VCRRequestsHTTPConnection)
        assert issubclass(VerifiedHTTPSConnection, VCRRequestsHTTPSConnection)
        HTTPConnection('hostname.does.not.matter')
        VerifiedHTTPSConnection('hostname.does.not.matter')


def test_boto3_without_vcr():
    s3_resource = boto3.resource('s3')
    b = s3_resource.Bucket(bucket)
    b.put_object(Key=key, Body=content)

    # retrieve content to check it
    o = s3_resource.Object(bucket, key).get()

    # decode for python3
    assert content == o['Body'].read().decode('utf-8')


def test_boto_medium_difficulty(tmpdir):
    s3_resource = boto3.resource('s3')
    b = s3_resource.Bucket(bucket)
    with vcr.use_cassette(str(tmpdir.join('boto3-medium.yml'))):
        b.put_object(Key=key, Body=content)
        o = s3_resource.Object(bucket, key).get()
        assert content == o['Body'].read().decode('utf-8')

    with vcr.use_cassette(str(tmpdir.join('boto3-medium.yml'))) as cass:
        b.put_object(Key=key, Body=content)
        o = s3_resource.Object(bucket, key).get()
        assert content == o['Body'].read().decode('utf-8')
        assert cass.all_played


def test_boto_hardcore_mode(tmpdir):
    with vcr.use_cassette(str(tmpdir.join('boto3-hardcore.yml'))):
        s3_resource = boto3.resource('s3')
        b = s3_resource.Bucket(bucket)
        b.put_object(Key=key, Body=content)
        o = s3_resource.Object(bucket, key).get()
        assert content == o['Body'].read().decode('utf-8')

    with vcr.use_cassette(str(tmpdir.join('boto3-hardcore.yml'))) as cass:
        s3_resource = boto3.resource('s3')
        b = s3_resource.Bucket(bucket)
        b.put_object(Key=key, Body=content)
        o = s3_resource.Object(bucket, key).get()
        assert content == o['Body'].read().decode('utf-8')
        assert cass.all_played



from six.moves.urllib.request import urlopen
import socket
from contextlib import contextmanager
import vcr


@contextmanager
def overridden_dns(overrides):
    """
    Monkeypatch socket.getaddrinfo() to override DNS lookups (name will resolve
    to address)
    """
    real_getaddrinfo = socket.getaddrinfo

    def fake_getaddrinfo(*args, **kwargs):
        if args[0] in overrides:
            address = overrides[args[0]]
            return [(2, 1, 6, '', (address, args[1]))]
        return real_getaddrinfo(*args, **kwargs)
    socket.getaddrinfo = fake_getaddrinfo
    yield
    socket.getaddrinfo = real_getaddrinfo


def test_ignore_localhost(tmpdir, httpbin):
    with overridden_dns({'httpbin.org': '127.0.0.1'}):
        cass_file = str(tmpdir.join('filter_qs.yaml'))
        with vcr.use_cassette(cass_file, ignore_localhost=True) as cass:
            urlopen('http://localhost:{0}/'.format(httpbin.port))
            assert len(cass) == 0
            urlopen('http://httpbin.org:{0}/'.format(httpbin.port))
            assert len(cass) == 1


def test_ignore_httpbin(tmpdir, httpbin):
    with overridden_dns({'httpbin.org': '127.0.0.1'}):
        cass_file = str(tmpdir.join('filter_qs.yaml'))
        with vcr.use_cassette(
            cass_file,
            ignore_hosts=['httpbin.org']
        ) as cass:
            urlopen('http://httpbin.org:{0}/'.format(httpbin.port))
            assert len(cass) == 0
            urlopen('http://localhost:{0}/'.format(httpbin.port))
            assert len(cass) == 1


def test_ignore_localhost_and_httpbin(tmpdir, httpbin):
    with overridden_dns({'httpbin.org': '127.0.0.1'}):
        cass_file = str(tmpdir.join('filter_qs.yaml'))
        with vcr.use_cassette(
            cass_file,
            ignore_hosts=['httpbin.org'],
            ignore_localhost=True
        ) as cass:
            urlopen('http://httpbin.org:{0}'.format(httpbin.port))
            urlopen('http://localhost:{0}'.format(httpbin.port))
            assert len(cass) == 0


def test_ignore_localhost_twice(tmpdir, httpbin):
    with overridden_dns({'httpbin.org': '127.0.0.1'}):
        cass_file = str(tmpdir.join('filter_qs.yaml'))
        with vcr.use_cassette(cass_file, ignore_localhost=True) as cass:
            urlopen('http://localhost:{0}'.format(httpbin.port))
            assert len(cass) == 0
            urlopen('http://httpbin.org:{0}'.format(httpbin.port))
            assert len(cass) == 1
        with vcr.use_cassette(cass_file, ignore_localhost=True) as cass:
            assert len(cass) == 1
            urlopen('http://localhost:{0}'.format(httpbin.port))
            urlopen('http://httpbin.org:{0}'.format(httpbin.port))
            assert len(cass) == 1



import vcr
import pytest
from six.moves.urllib.request import urlopen


DEFAULT_URI = 'http://httpbin.org/get?p1=q1&p2=q2'  # base uri for testing


def _replace_httpbin(uri, httpbin, httpbin_secure):
    return uri.replace('http://httpbin.org', httpbin.url).replace('https://httpbin.org', httpbin_secure.url)


@pytest.fixture
def cassette(tmpdir, httpbin, httpbin_secure):
    """
    Helper fixture used to prepare the cassete
    returns path to the recorded cassette
    """
    default_uri = _replace_httpbin(DEFAULT_URI, httpbin, httpbin_secure)

    cassette_path = str(tmpdir.join('test.yml'))
    with vcr.use_cassette(cassette_path, record_mode='all'):
        urlopen(default_uri)
    return cassette_path


@pytest.mark.parametrize("matcher, matching_uri, not_matching_uri", [
    ('uri',
     'http://httpbin.org/get?p1=q1&p2=q2',
     'http://httpbin.org/get?p2=q2&p1=q1'),
    ('scheme',
     'http://google.com/post?a=b',
     'https://httpbin.org/get?p1=q1&p2=q2'),
    ('host',
     'https://httpbin.org/post?a=b',
     'http://google.com/get?p1=q1&p2=q2'),
    ('path',
     'https://google.com/get?a=b',
     'http://httpbin.org/post?p1=q1&p2=q2'),
    ('query',
     'https://google.com/get?p2=q2&p1=q1',
     'http://httpbin.org/get?p1=q1&a=b')
])
def test_matchers(httpbin, httpbin_secure, cassette, matcher, matching_uri, not_matching_uri):

    matching_uri = _replace_httpbin(matching_uri, httpbin, httpbin_secure)
    not_matching_uri = _replace_httpbin(not_matching_uri, httpbin, httpbin_secure)
    default_uri = _replace_httpbin(DEFAULT_URI, httpbin, httpbin_secure)

    # play cassette with default uri
    with vcr.use_cassette(cassette, match_on=[matcher]) as cass:
        urlopen(default_uri)
        assert cass.play_count == 1

    # play cassette with matching on uri
    with vcr.use_cassette(cassette, match_on=[matcher]) as cass:
        urlopen(matching_uri)
        assert cass.play_count == 1

    # play cassette with not matching on uri, it should fail
    with pytest.raises(vcr.errors.CannotOverwriteExistingCassetteException):
        with vcr.use_cassette(cassette, match_on=[matcher]) as cass:
            urlopen(not_matching_uri)


def test_method_matcher(cassette, httpbin, httpbin_secure):
    default_uri = _replace_httpbin(DEFAULT_URI, httpbin, httpbin_secure)

    # play cassette with matching on method
    with vcr.use_cassette(cassette, match_on=['method']) as cass:
        urlopen('https://google.com/get?a=b')
        assert cass.play_count == 1

    # should fail if method does not match
    with pytest.raises(vcr.errors.CannotOverwriteExistingCassetteException):
        with vcr.use_cassette(cassette, match_on=['method']) as cass:
            # is a POST request
            urlopen(default_uri, data=b'')


@pytest.mark.parametrize("uri", [
    DEFAULT_URI,
    'http://httpbin.org/get?p2=q2&p1=q1',
    'http://httpbin.org/get?p2=q2&p1=q1',
])
def test_default_matcher_matches(cassette, uri, httpbin, httpbin_secure):

    uri = _replace_httpbin(uri, httpbin, httpbin_secure)

    with vcr.use_cassette(cassette) as cass:
        urlopen(uri)
        assert cass.play_count == 1


@pytest.mark.parametrize("uri", [
    'https://httpbin.org/get?p1=q1&p2=q2',
    'http://google.com/get?p1=q1&p2=q2',
    'http://httpbin.org/post?p1=q1&p2=q2',
    'http://httpbin.org/get?p1=q1&a=b'
])
def test_default_matcher_does_not_match(cassette, uri, httpbin, httpbin_secure):
    uri = _replace_httpbin(uri, httpbin, httpbin_secure)
    with pytest.raises(vcr.errors.CannotOverwriteExistingCassetteException):
        with vcr.use_cassette(cassette):
            urlopen(uri)


def test_default_matcher_does_not_match_on_method(cassette, httpbin, httpbin_secure):
    default_uri = _replace_httpbin(DEFAULT_URI, httpbin, httpbin_secure)
    with pytest.raises(vcr.errors.CannotOverwriteExistingCassetteException):
        with vcr.use_cassette(cassette):
            # is a POST request
            urlopen(default_uri, data=b'')



from vcr.stubs import VCRHTTPSConnection


class TestVCRConnection(object):

    def test_setting_of_attributes_get_propogated_to_real_connection(self):
        vcr_connection = VCRHTTPSConnection('www.examplehost.com')
        vcr_connection.ssl_version = 'example_ssl_version'
        assert vcr_connection.real_connection.ssl_version == 'example_ssl_version'



import os


class FilesystemPersister(object):
    @classmethod
    def write(cls, cassette_path, data):
        dirname, filename = os.path.split(cassette_path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(cassette_path, 'w') as f:
            f.write(data)



try:
    from unittest import mock
except ImportError:
    import mock

try:
    import contextlib
except ImportError:
    import contextlib2 as contextlib
else:
    if not hasattr(contextlib, 'ExitStack'):
        import contextlib2 as contextlib

import collections
if not hasattr(collections, 'Counter'):
    import backport_collections as collections

__all__ = ['mock', 'contextlib', 'collections']



# -*- coding: utf-8 -*-
'''Basic tests about save behavior'''

# External imports
import os
import time
from six.moves.urllib.request import urlopen

# Internal imports
import vcr


def test_disk_saver_nowrite(tmpdir, httpbin):
    '''
    Ensure that when you close a cassette without changing it it doesn't
    rewrite the file
    '''
    fname = str(tmpdir.join('synopsis.yaml'))
    with vcr.use_cassette(fname) as cass:
        urlopen(httpbin.url).read()
        assert cass.play_count == 0
    last_mod = os.path.getmtime(fname)

    with vcr.use_cassette(fname) as cass:
        urlopen(httpbin.url).read()
        assert cass.play_count == 1
        assert cass.dirty is False
    last_mod2 = os.path.getmtime(fname)

    assert last_mod == last_mod2


def test_disk_saver_write(tmpdir, httpbin):
    '''
    Ensure that when you close a cassette after changing it it does
    rewrite the file
    '''
    fname = str(tmpdir.join('synopsis.yaml'))
    with vcr.use_cassette(fname) as cass:
        urlopen(httpbin.url).read()
        assert cass.play_count == 0
    last_mod = os.path.getmtime(fname)

    # Make sure at least 1 second passes, otherwise sometimes
    # the mtime doesn't change
    time.sleep(1)

    with vcr.use_cassette(fname, record_mode='any') as cass:
        urlopen(httpbin.url).read()
        urlopen(httpbin.url + '/get').read()
        assert cass.play_count == 1
        assert cass.dirty
    last_mod2 = os.path.getmtime(fname)

    assert last_mod != last_mod2



import vcr


class MockSerializer(object):
    def __init__(self):
        self.serialize_count = 0
        self.deserialize_count = 0
        self.load_args = None

    def deserialize(self, cassette_string):
        self.serialize_count += 1
        self.cassette_string = cassette_string
        return {'interactions': []}

    def serialize(self, cassette_dict):
        self.deserialize_count += 1
        return ""


def test_registered_serializer(tmpdir):
    ms = MockSerializer()
    my_vcr = vcr.VCR()
    my_vcr.register_serializer('mock', ms)
    tmpdir.join('test.mock').write('test_data')
    with my_vcr.use_cassette(str(tmpdir.join('test.mock')), serializer='mock'):
        # Serializer deserialized once
        assert ms.serialize_count == 1
        # and serialized the test data string
        assert ms.cassette_string == 'test_data'
        # and hasn't serialized yet
        assert ms.deserialize_count == 0

    assert ms.serialize_count == 1



import base64
import pytest
from six.moves.urllib.request import urlopen, Request
from six.moves.urllib.parse import urlencode
from six.moves.urllib.error import HTTPError
import vcr
import json
from assertions import assert_cassette_has_one_response, assert_is_json


def _request_with_auth(url, username, password):
    request = Request(url)
    base64string = base64.b64encode(
        username.encode('ascii') + b':' + password.encode('ascii')
    )
    request.add_header(b"Authorization", b"Basic " + base64string)
    return urlopen(request)


def _find_header(cassette, header):
    return any(header in request.headers for request in cassette.requests)


def test_filter_basic_auth(tmpdir, httpbin):
    url = httpbin.url + '/basic-auth/user/passwd'
    cass_file = str(tmpdir.join('basic_auth_filter.yaml'))
    my_vcr = vcr.VCR(match_on=['uri', 'method', 'headers'])
    # 2 requests, one with auth failure and one with auth success
    with my_vcr.use_cassette(cass_file, filter_headers=['authorization']):
        with pytest.raises(HTTPError):
            resp = _request_with_auth(url, 'user', 'wrongpasswd')
            assert resp.getcode() == 401
        resp = _request_with_auth(url, 'user', 'passwd')
        assert resp.getcode() == 200
    # make same 2 requests, this time both served from cassette.
    with my_vcr.use_cassette(cass_file, filter_headers=['authorization']) as cass:
        with pytest.raises(HTTPError):
            resp = _request_with_auth(url, 'user', 'wrongpasswd')
            assert resp.getcode() == 401
        resp = _request_with_auth(url, 'user', 'passwd')
        assert resp.getcode() == 200
        # authorization header should not have been recorded
        assert not _find_header(cass, 'authorization')
        assert len(cass) == 2


def test_filter_querystring(tmpdir, httpbin):
    url = httpbin.url + '/?foo=bar'
    cass_file = str(tmpdir.join('filter_qs.yaml'))
    with vcr.use_cassette(cass_file, filter_query_parameters=['foo']):
        urlopen(url)
    with vcr.use_cassette(cass_file, filter_query_parameters=['foo']) as cass:
        urlopen(url)
        assert 'foo' not in cass.requests[0].url


def test_filter_post_data(tmpdir, httpbin):
    url = httpbin.url + '/post'
    data = urlencode({'id': 'secret', 'foo': 'bar'}).encode('utf-8')
    cass_file = str(tmpdir.join('filter_pd.yaml'))
    with vcr.use_cassette(cass_file, filter_post_data_parameters=['id']):
        urlopen(url, data)
    with vcr.use_cassette(cass_file, filter_post_data_parameters=['id']) as cass:
        assert b'id=secret' not in cass.requests[0].body


def test_filter_json_post_data(tmpdir, httpbin):
    data = json.dumps({'id': 'secret', 'foo': 'bar'}).encode('utf-8')
    request = Request(httpbin.url + '/post', data=data)
    request.add_header('Content-Type', 'application/json')

    cass_file = str(tmpdir.join('filter_jpd.yaml'))
    with vcr.use_cassette(cass_file, filter_post_data_parameters=['id']):
        urlopen(request)
    with vcr.use_cassette(cass_file, filter_post_data_parameters=['id']) as cass:
        assert b'"id": "secret"' not in cass.requests[0].body


def test_filter_callback(tmpdir, httpbin):
    url = httpbin.url + '/get'
    cass_file = str(tmpdir.join('basic_auth_filter.yaml'))

    def before_record_cb(request):
        if request.path != '/get':
            return request

    # Test the legacy keyword.
    my_vcr = vcr.VCR(before_record=before_record_cb)
    with my_vcr.use_cassette(cass_file, filter_headers=['authorization']) as cass:
        urlopen(url)
        assert len(cass) == 0

    my_vcr = vcr.VCR(before_record_request=before_record_cb)
    with my_vcr.use_cassette(cass_file, filter_headers=['authorization']) as cass:
        urlopen(url)
        assert len(cass) == 0


def test_decompress_gzip(tmpdir, httpbin):
    url = httpbin.url + '/gzip'
    request = Request(url, headers={'Accept-Encoding': ['gzip, deflate']})
    cass_file = str(tmpdir.join('gzip_response.yaml'))
    with vcr.use_cassette(cass_file, decode_compressed_response=True):
        urlopen(request)
    with vcr.use_cassette(cass_file) as cass:
        decoded_response = urlopen(url).read()
        assert_cassette_has_one_response(cass)
    assert_is_json(decoded_response)


def test_decompress_deflate(tmpdir, httpbin):
    url = httpbin.url + '/deflate'
    request = Request(url, headers={'Accept-Encoding': ['gzip, deflate']})
    cass_file = str(tmpdir.join('deflate_response.yaml'))
    with vcr.use_cassette(cass_file, decode_compressed_response=True):
        urlopen(request)
    with vcr.use_cassette(cass_file) as cass:
        decoded_response = urlopen(url).read()
        assert_cassette_has_one_response(cass)
    assert_is_json(decoded_response)


def test_decompress_regular(tmpdir, httpbin):
    """Test that it doesn't try to decompress content that isn't compressed"""
    url = httpbin.url + '/get'
    cass_file = str(tmpdir.join('noncompressed_response.yaml'))
    with vcr.use_cassette(cass_file, decode_compressed_response=True):
        urlopen(url)
    with vcr.use_cassette(cass_file) as cass:
        resp = urlopen(url).read()
        assert_cassette_has_one_response(cass)
    assert_is_json(resp)



import pytest
import vcr
from six.moves.urllib.request import urlopen


def test_once_record_mode(tmpdir, httpbin):
    testfile = str(tmpdir.join('recordmode.yml'))
    with vcr.use_cassette(testfile, record_mode="once"):
        # cassette file doesn't exist, so create.
        urlopen(httpbin.url).read()

    with vcr.use_cassette(testfile, record_mode="once"):
        # make the same request again
        urlopen(httpbin.url).read()

        # the first time, it's played from the cassette.
        # but, try to access something else from the same cassette, and an
        # exception is raised.
        with pytest.raises(Exception):
            urlopen(httpbin.url + '/get').read()


def test_once_record_mode_two_times(tmpdir, httpbin):
    testfile = str(tmpdir.join('recordmode.yml'))
    with vcr.use_cassette(testfile, record_mode="once"):
        # get two of the same file
        urlopen(httpbin.url).read()
        urlopen(httpbin.url).read()

    with vcr.use_cassette(testfile, record_mode="once"):
        # do it again
        urlopen(httpbin.url).read()
        urlopen(httpbin.url).read()


def test_once_mode_three_times(tmpdir, httpbin):
    testfile = str(tmpdir.join('recordmode.yml'))
    with vcr.use_cassette(testfile, record_mode="once"):
        # get three of the same file
        urlopen(httpbin.url).read()
        urlopen(httpbin.url).read()
        urlopen(httpbin.url).read()


def test_new_episodes_record_mode(tmpdir, httpbin):
    testfile = str(tmpdir.join('recordmode.yml'))

    with vcr.use_cassette(testfile, record_mode="new_episodes"):
        # cassette file doesn't exist, so create.
        urlopen(httpbin.url).read()

    with vcr.use_cassette(testfile, record_mode="new_episodes") as cass:
        # make the same request again
        urlopen(httpbin.url).read()

        # all responses have been played
        assert cass.all_played

        # in the "new_episodes" record mode, we can add more requests to
        # a cassette without repurcussions.
        urlopen(httpbin.url + '/get').read()

        # one of the responses has been played
        assert cass.play_count == 1

        # not all responses have been played
        assert not cass.all_played

    with vcr.use_cassette(testfile, record_mode="new_episodes") as cass:
        # the cassette should now have 2 responses
        assert len(cass.responses) == 2


def test_new_episodes_record_mode_two_times(tmpdir, httpbin):
    testfile = str(tmpdir.join('recordmode.yml'))
    url = httpbin.url + '/bytes/1024'
    with vcr.use_cassette(testfile, record_mode="new_episodes"):
        # cassette file doesn't exist, so create.
        original_first_response = urlopen(url).read()

    with vcr.use_cassette(testfile, record_mode="new_episodes"):
        # make the same request again
        assert urlopen(url).read() == original_first_response

        # in the "new_episodes" record mode, we can add the same request
        # to the cassette without repercussions
        original_second_response = urlopen(url).read()

    with vcr.use_cassette(testfile, record_mode="once"):
        # make the same request again
        assert urlopen(url).read() == original_first_response
        assert urlopen(url).read() == original_second_response
        # now that we are back in once mode, this should raise
        # an error.
        with pytest.raises(Exception):
            urlopen(url).read()


def test_all_record_mode(tmpdir, httpbin):
    testfile = str(tmpdir.join('recordmode.yml'))

    with vcr.use_cassette(testfile, record_mode="all"):
        # cassette file doesn't exist, so create.
        urlopen(httpbin.url).read()

    with vcr.use_cassette(testfile, record_mode="all") as cass:
        # make the same request again
        urlopen(httpbin.url).read()

        # in the "all" record mode, we can add more requests to
        # a cassette without repurcussions.
        urlopen(httpbin.url + '/get').read()

        # The cassette was never actually played, even though it existed.
        # that's because, in "all" mode, the requests all go directly to
        # the source and bypass the cassette.
        assert cass.play_count == 0


def test_none_record_mode(tmpdir, httpbin):
    # Cassette file doesn't exist, yet we are trying to make a request.
    # raise hell.
    testfile = str(tmpdir.join('recordmode.yml'))
    with vcr.use_cassette(testfile, record_mode="none"):
        with pytest.raises(Exception):
            urlopen(httpbin.url).read()


def test_none_record_mode_with_existing_cassette(tmpdir, httpbin):
    # create a cassette file
    testfile = str(tmpdir.join('recordmode.yml'))

    with vcr.use_cassette(testfile, record_mode="all"):
        urlopen(httpbin.url).read()

    # play from cassette file
    with vcr.use_cassette(testfile, record_mode="none") as cass:
        urlopen(httpbin.url).read()
        assert cass.play_count == 1
        # but if I try to hit the net, raise an exception.
        with pytest.raises(Exception):
            urlopen(httpbin.url + '/get').read()



#!/usr/bin/env python

import sys
import logging

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import pkg_resources

long_description = open('README.rst', 'r').read()


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


install_requires = ['PyYAML', 'wrapt', 'six>=1.5']


extras_require = {
    ':python_version in "2.4, 2.5, 2.6"':
        ['contextlib2', 'backport_collections', 'mock'],
    ':python_version in "2.7, 3.1, 3.2"': ['contextlib2', 'mock'],
}


try:
    if 'bdist_wheel' not in sys.argv:
        for key, value in extras_require.items():
            if key.startswith(':') and pkg_resources.evaluate_marker(key[1:]):
                install_requires.extend(value)
except Exception:
    logging.getLogger(__name__).exception(
        'Something went wrong calculating platform specific dependencies, so '
        "you're getting them all!"
    )
    for key, value in extras_require.items():
        if key.startswith(':'):
            install_requires.extend(value)


setup(
    name='vcrpy',
    version='1.10.0',
    description=(
        "Automatically mock your HTTP interactions to simplify and "
        "speed up testing"
    ),
    long_description=long_description,
    author='Kevin McCarthy',
    author_email='me@kevinmccarthy.org',
    url='https://github.com/kevin1024/vcrpy',
    packages=find_packages(exclude=("tests*",)),
    install_requires=install_requires,
    extras_require=extras_require,
    license='MIT',
    tests_require=['pytest', 'mock', 'pytest-httpbin'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: MIT License',
    ]
)



import json


def assert_cassette_empty(cass):
    assert len(cass) == 0
    assert cass.play_count == 0


def assert_cassette_has_one_response(cass):
    assert len(cass) == 1
    assert cass.play_count == 1


def assert_is_json(a_string):
    try:
        json.loads(a_string.decode('utf-8'))
    except Exception:
        assert False
    assert True



from six import BytesIO
from vcr.filters import (
    remove_headers, replace_headers,
    remove_query_parameters, replace_query_parameters,
    remove_post_data_parameters, replace_post_data_parameters,
    decode_response
)
from vcr.compat import mock
from vcr.request import Request
import gzip
import json
import zlib


def test_replace_headers():
    # This tests all of:
    #   1. keeping a header
    #   2. removing a header
    #   3. replacing a header
    #   4. replacing a header using a callable
    #   5. removing a header using a callable
    #   6. replacing a header that doesn't exist
    headers = {
        'one': ['keep'],
        'two': ['lose'],
        'three': ['change'],
        'four': ['shout'],
        'five': ['whisper'],
    }
    request = Request('GET', 'http://google.com', '', headers)
    replace_headers(request, [
        ('two', None),
        ('three', 'tada'),
        ('four', lambda key, value, request: value.upper()),
        ('five', lambda key, value, request: None),
        ('six', 'doesntexist'),
    ])
    assert request.headers == {
        'one': 'keep',
        'three': 'tada',
        'four': 'SHOUT',
    }


def test_replace_headers_empty():
    headers = {'hello': 'goodbye', 'secret': 'header'}
    request = Request('GET', 'http://google.com', '', headers)
    replace_headers(request, [])
    assert request.headers == headers


def test_replace_headers_callable():
    # This goes beyond test_replace_headers() to ensure that the callable
    # receives the expected arguments.
    headers = {'hey': 'there'}
    request = Request('GET', 'http://google.com', '', headers)
    callme = mock.Mock(return_value='ho')
    replace_headers(request, [('hey', callme)])
    assert request.headers == {'hey': 'ho'}
    assert callme.call_args == ((), {'request': request,
                                     'key': 'hey',
                                     'value': 'there'})


def test_remove_headers():
    # Test the backward-compatible API wrapper.
    headers = {'hello': ['goodbye'], 'secret': ['header']}
    request = Request('GET', 'http://google.com', '', headers)
    remove_headers(request, ['secret'])
    assert request.headers == {'hello': 'goodbye'}


def test_replace_query_parameters():
    # This tests all of:
    #   1. keeping a parameter
    #   2. removing a parameter
    #   3. replacing a parameter
    #   4. replacing a parameter using a callable
    #   5. removing a parameter using a callable
    #   6. replacing a parameter that doesn't exist
    uri = 'http://g.com/?one=keep&two=lose&three=change&four=shout&five=whisper'
    request = Request('GET', uri, '', {})
    replace_query_parameters(request, [
        ('two', None),
        ('three', 'tada'),
        ('four', lambda key, value, request: value.upper()),
        ('five', lambda key, value, request: None),
        ('six', 'doesntexist'),
    ])
    assert request.query == [
        ('four', 'SHOUT'),
        ('one', 'keep'),
        ('three', 'tada'),
    ]


def test_remove_all_query_parameters():
    uri = 'http://g.com/?q=cowboys&w=1'
    request = Request('GET', uri, '', {})
    replace_query_parameters(request, [('w', None), ('q', None)])
    assert request.uri == 'http://g.com/'


def test_replace_query_parameters_callable():
    # This goes beyond test_replace_query_parameters() to ensure that the
    # callable receives the expected arguments.
    uri = 'http://g.com/?hey=there'
    request = Request('GET', uri, '', {})
    callme = mock.Mock(return_value='ho')
    replace_query_parameters(request, [('hey', callme)])
    assert request.uri == 'http://g.com/?hey=ho'
    assert callme.call_args == ((), {'request': request,
                                     'key': 'hey',
                                     'value': 'there'})


def test_remove_query_parameters():
    # Test the backward-compatible API wrapper.
    uri = 'http://g.com/?q=cowboys&w=1'
    request = Request('GET', uri, '', {})
    remove_query_parameters(request, ['w'])
    assert request.uri == 'http://g.com/?q=cowboys'


def test_replace_post_data_parameters():
    # This tests all of:
    #   1. keeping a parameter
    #   2. removing a parameter
    #   3. replacing a parameter
    #   4. replacing a parameter using a callable
    #   5. removing a parameter using a callable
    #   6. replacing a parameter that doesn't exist
    body = b'one=keep&two=lose&three=change&four=shout&five=whisper'
    request = Request('POST', 'http://google.com', body, {})
    replace_post_data_parameters(request, [
        ('two', None),
        ('three', 'tada'),
        ('four', lambda key, value, request: value.upper()),
        ('five', lambda key, value, request: None),
        ('six', 'doesntexist'),
    ])
    assert request.body == b'one=keep&three=tada&four=SHOUT'


def test_remove_post_data_parameters():
    # Test the backward-compatible API wrapper.
    body = b'id=secret&foo=bar'
    request = Request('POST', 'http://google.com', body, {})
    remove_post_data_parameters(request, ['id'])
    assert request.body == b'foo=bar'


def test_preserve_multiple_post_data_parameters():
    body = b'id=secret&foo=bar&foo=baz'
    request = Request('POST', 'http://google.com', body, {})
    replace_post_data_parameters(request, [('id', None)])
    assert request.body == b'foo=bar&foo=baz'


def test_remove_all_post_data_parameters():
    body = b'id=secret&foo=bar'
    request = Request('POST', 'http://google.com', body, {})
    replace_post_data_parameters(request, [('id', None), ('foo', None)])
    assert request.body == b''


def test_replace_json_post_data_parameters():
    # This tests all of:
    #   1. keeping a parameter
    #   2. removing a parameter
    #   3. replacing a parameter
    #   4. replacing a parameter using a callable
    #   5. removing a parameter using a callable
    #   6. replacing a parameter that doesn't exist
    body = b'{"one": "keep", "two": "lose", "three": "change", "four": "shout", "five": "whisper"}'
    request = Request('POST', 'http://google.com', body, {})
    request.headers['Content-Type'] = 'application/json'
    replace_post_data_parameters(request, [
        ('two', None),
        ('three', 'tada'),
        ('four', lambda key, value, request: value.upper()),
        ('five', lambda key, value, request: None),
        ('six', 'doesntexist'),
    ])
    request_data = json.loads(request.body.decode('utf-8'))
    expected_data = json.loads('{"one": "keep", "three": "tada", "four": "SHOUT"}')
    assert request_data == expected_data


def test_remove_json_post_data_parameters():
    # Test the backward-compatible API wrapper.
    body = b'{"id": "secret", "foo": "bar", "baz": "qux"}'
    request = Request('POST', 'http://google.com', body, {})
    request.headers['Content-Type'] = 'application/json'
    remove_post_data_parameters(request, ['id'])
    request_body_json = json.loads(request.body.decode('utf-8'))
    expected_json = json.loads(b'{"foo": "bar", "baz": "qux"}'.decode('utf-8'))
    assert request_body_json == expected_json


def test_remove_all_json_post_data_parameters():
    body = b'{"id": "secret", "foo": "bar"}'
    request = Request('POST', 'http://google.com', body, {})
    request.headers['Content-Type'] = 'application/json'
    replace_post_data_parameters(request, [('id', None), ('foo', None)])
    assert request.body == b'{}'


def test_decode_response_uncompressed():
    recorded_response = {
        "status": {
            "message": "OK",
            "code": 200
        },
        "headers": {
            "content-length": ["10806"],
            "date": ["Fri, 24 Oct 2014 18:35:37 GMT"],
            "content-type": ["text/html; charset=utf-8"],
        },
        "body": {
            "string": b""
        }
    }
    assert decode_response(recorded_response) == recorded_response


def test_decode_response_deflate():
    body = b'deflate message'
    deflate_response = {
        'body': {'string': zlib.compress(body)},
        'headers': {
            'access-control-allow-credentials': ['true'],
            'access-control-allow-origin': ['*'],
            'connection': ['keep-alive'],
            'content-encoding': ['deflate'],
            'content-length': ['177'],
            'content-type': ['application/json'],
            'date': ['Wed, 02 Dec 2015 19:44:32 GMT'],
            'server': ['nginx']
        },
        'status': {'code': 200, 'message': 'OK'}
    }
    decoded_response = decode_response(deflate_response)
    assert decoded_response['body']['string'] == body
    assert decoded_response['headers']['content-length'] == [str(len(body))]


def test_decode_response_gzip():
    body = b'gzip message'

    buf = BytesIO()
    f = gzip.GzipFile('a', fileobj=buf, mode='wb')
    f.write(body)
    f.close()

    compressed_body = buf.getvalue()
    buf.close()
    gzip_response = {
        'body': {'string': compressed_body},
        'headers': {
            'access-control-allow-credentials': ['true'],
            'access-control-allow-origin': ['*'],
            'connection': ['keep-alive'],
            'content-encoding': ['gzip'],
            'content-length': ['177'],
            'content-type': ['application/json'],
            'date': ['Wed, 02 Dec 2015 19:44:32 GMT'],
            'server': ['nginx']
        },
        'status': {'code': 200, 'message': 'OK'}
    }
    decoded_response = decode_response(gzip_response)
    assert decoded_response['body']['string'] == body
    assert decoded_response['headers']['content-length'] == [str(len(body))]



'''Stubs for aiohttp HTTP clients'''
from __future__ import absolute_import

import asyncio
import functools
import json

from aiohttp import ClientResponse

from vcr.request import Request


class MockClientResponse(ClientResponse):
    # TODO: get encoding from header
    @asyncio.coroutine
    def json(self, *, encoding='utf-8', loads=json.loads):  # NOQA: E999
        return loads(self.content.decode(encoding))

    @asyncio.coroutine
    def text(self, encoding='utf-8'):
        return self.content.decode(encoding)

    @asyncio.coroutine
    def release(self):
        pass


def vcr_request(cassette, real_request):

    @functools.wraps(real_request)
    @asyncio.coroutine
    def new_request(self, method, url, **kwargs):
        headers = kwargs.get('headers')
        headers = self._prepare_headers(headers)
        data = kwargs.get('data')

        vcr_request = Request(method, url, data, headers)

        if cassette.can_play_response_for(vcr_request):
            vcr_response = cassette.play_response(vcr_request)

            response = MockClientResponse(method, vcr_response.get('url'))
            response.status = vcr_response['status']['code']
            response.content = vcr_response['body']['string']
            response.reason = vcr_response['status']['message']
            response.headers = vcr_response['headers']

            response.close()
            return response

        if cassette.write_protected and cassette.filter_request(vcr_request):
            response = MockClientResponse(method, url)
            response.status = 599
            msg = ("No match for the request {!r} was found. Can't overwrite "
                   "existing cassette {!r} in your current record mode {!r}.")
            msg = msg.format(vcr_request, cassette._path, cassette.record_mode)
            response.content = msg.encode()
            response.close()
            return response

        response = yield from real_request(self, method, url, **kwargs)  # NOQA: E999

        vcr_response = {
            'status': {
                'code': response.status,
                'message': response.reason,
            },
            'headers': dict(response.headers),
            'body': {'string': (yield from response.text())},  # NOQA: E999
            'url': response.url,
        }
        cassette.append(vcr_request, vcr_response)

        return response

    return new_request



import os
import json
import pytest
import vcr
from six.moves.urllib.request import urlopen


def test_set_serializer_default_config(tmpdir, httpbin):
    my_vcr = vcr.VCR(serializer='json')

    with my_vcr.use_cassette(str(tmpdir.join('test.json'))):
        assert my_vcr.serializer == 'json'
        urlopen(httpbin.url + '/get')

    with open(str(tmpdir.join('test.json'))) as f:
        assert json.loads(f.read())


def test_default_set_cassette_library_dir(tmpdir, httpbin):
    my_vcr = vcr.VCR(cassette_library_dir=str(tmpdir.join('subdir')))

    with my_vcr.use_cassette('test.json'):
        urlopen(httpbin.url + '/get')

    assert os.path.exists(str(tmpdir.join('subdir').join('test.json')))


def test_override_set_cassette_library_dir(tmpdir, httpbin):
    my_vcr = vcr.VCR(cassette_library_dir=str(tmpdir.join('subdir')))

    cld = str(tmpdir.join('subdir2'))

    with my_vcr.use_cassette('test.json', cassette_library_dir=cld):
        urlopen(httpbin.url + '/get')

    assert os.path.exists(str(tmpdir.join('subdir2').join('test.json')))
    assert not os.path.exists(str(tmpdir.join('subdir').join('test.json')))


def test_override_match_on(tmpdir, httpbin):
    my_vcr = vcr.VCR(match_on=['method'])

    with my_vcr.use_cassette(str(tmpdir.join('test.json'))):
        urlopen(httpbin.url)

    with my_vcr.use_cassette(str(tmpdir.join('test.json'))) as cass:
        urlopen(httpbin.url + '/get')

    assert len(cass) == 1
    assert cass.play_count == 1


def test_missing_matcher():
    my_vcr = vcr.VCR()
    my_vcr.register_matcher("awesome", object)
    with pytest.raises(KeyError):
        with my_vcr.use_cassette("test.yaml", match_on=['notawesome']):
            pass
